##################### generated by xml-casa (v2) from concateovsa.xml ###############
##################### cdf0275a4ed32b7fdcd90e799b22f760 ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from suncasatasks import concateovsa as _concateovsa_t
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate

class _concateovsa:
    """
    concateovsa ---- Concatenate several EOVSA visibility data sets.

    
    This is a EOVSA version of CASA concat task.
    
    The list of data sets given in the vis argument are chronologically concatenated
    into an output data set in concatvis, i.e. the data sets in vis are first ordered
    by the time of their earliest integration and then concatenated.
    
    If there are fields whose direction agrees within the direction tolerance
    (parameter dirtol), the actual direction in the resulting, merged output field
    will be the one from the chronologically first input MS.
    
    If concatvis already exists (e.g., it is the same as the first input data set),
    then the other input data sets will be appended to the concatvis data set.
    There is no limit to the number of input data sets.
    
    If none of the input data sets have any scratch columns (model and corrected
    columns), none are created in the concatvis. Otherwise these columns are
    created on output and initialized to their default value (1 in model column,
    data in corrected column) for those data with no input columns.
    
    Spectral windows for each data set with the same chanelization, and within a
    specified frequency tolerance of another data set will be combined into one
    spectral window.
    
    A field position in one data set that is within a specified direction tolerance
    of another field position in any other data set will be combined into one
    field. The field names need not be the same---only their position is used.
    
    Each appended dataset is assigned a new observation id (provided the entries
    in the observation table are indeed different).
    
    Keyword arguments:
    vis -- Name of input visibility files to be combined
    default: none; example: vis = ['src2.ms','ngc5921.ms','ngc315.ms']
    concatvis -- Name of visibility file that will contain the concatenated data
    note: if this file exits on disk then the input files are
    added to this file. Otherwise the new file contains
    the concatenated data. Be careful here when concatenating to
    an existing file.
    default: none; example: concatvis='src2.ms'
    example: concatvis='outvis.ms'
    
    datacolumn -- Which data column to use for processing (case-insensitive).
    default: 'corrected'; example: datacolumn='data'
    options: 'data', 'corrected'.
    
    freqtol -- Frequency shift tolerance for considering data to be in the same
    spwid. The number of channels must also be the same.
    default: '' == 1 Hz
    example: freqtol='10MHz' will not combine spwid unless they are
    within 10 MHz.
    Note: This option is useful to combine spectral windows with very slight
    frequency differences caused by Doppler tracking, for example.
    
    dirtol -- Direction shift tolerance for considering data as the same field
    default: '' == 1 mas (milliarcsec)
    example: dirtol='1arcsec' will not combine data for a field unless
    their phase center differ by less than 1 arcsec. If the field names
    are different in the input data sets, the name in the output data
    set will be the first relevant data set in the list.
    
    respectname -- If true, fields with a different name are not merged even if their
    direction agrees (within dirtol)
    default: False
    
    timesort -- If true, the output visibility table will be sorted in time.
    default: false. Data in order as read in.
    example: timesort=true
    Note: There is no constraint on data that is simultaneously observed for
    more than one field; for example multi-source correlation of VLBA data.
    
    copypointing -- Make a proper copy of the POINTING subtable (can be time consuming).
    If False, the result is an empty POINTING table.
    default: True
    
    visweightscale -- The weights of the individual MSs will be scaled in the concatenated
    output MS by the factors in this list. SIGMA will be scaled by 1/sqrt(factor).
    Useful for handling heterogeneous arrays.
    Use plotms to inspect the "Wt" column as a reference for determining the scaling
    factors. See the cookbook for more details.
    example: [1.,3.,3.] - scale the weights of the second and third MS by a factor 3
    and the SIGMA column of these MS by a factor 1/sqrt(3).
    default: [] (empty list) - no scaling
    
    forcesingleephemfield -- By default, concat will only merge two ephemeris fields if
    the first ephemeris covers the time range of the second. Otherwise, two separate
    fields with separate ephemerides are placed in the output MS.
    In order to override this behaviour and make concat merge the non-overlapping
    or only partially overlapping input ephemerides, the name or id of the field
    in question needs to be placed into the list in parameter 'forcesingleephemfield'.
    example: ['Neptune'] - will make sure that there is only one joint ephemeris for
    field Neptune in the output MS
    default: '' - standard treatment of all ephemeris fields
    
    

    --------- parameter descriptions ---------------------------------------------

    vis                   Name of input visibility files to be concatenated
    concatvis             Name of output visibility file
    datacolumn            Which data column(s) to concatenate
    keep_orig_ms          If false, input vis files will be removed
    cols2rm               Columns in concatvis to be removed to slim the concatvis
    freqtol               Frequency shift tolerance for considering data as the same spwid
    dirtol                Direction shift tolerance for considering data as the same field
    respectname           If true, fields with a different name are not merged even if their direction agrees
    timesort              If true, sort by TIME in ascending order
    copypointing          Copy all rows of the POINTING table.
    visweightscale        List of the weight scaling factors to be applied to the individual MSs
    forcesingleephemfield make sure that there is only one joint ephemeris for every field in this list

    --------- examples -----------------------------------------------------------

    
    concateovsa(vis=['UDB20180102161402.ms','UDB20180102173518.ms'], concatvis='UDB20180102_allday.ms')
    will concatenate 'UDB20180102161402.ms' and 'UDB20180102173518.ms' into 'UDB20180102_allday.ms'
    


    """

    _info_group_ = """utility, manipulation"""
    _info_desc_ = """Concatenate several EOVSA visibility data sets."""

    __schema = {'vis': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'concatvis': {'type': 'cStr', 'coerce': _coerce.to_str}, 'datacolumn': {'type': 'cStr', 'coerce': _coerce.to_str, 'allowed': [ 'corrected', 'data', 'CORRECTED', 'DATA' ]}, 'keep_orig_ms': {'type': 'cBool'}, 'cols2rm': {'type': 'cStr', 'coerce': _coerce.to_str, 'allowed': [ 'model', 'corrected', 'CORRECTED', 'MODEL,CORRECTED', 'model,corrected', 'MODEL' ]}, 'freqtol': {'type': 'cVariant', 'coerce': [_coerce.to_variant]}, 'dirtol': {'type': 'cVariant', 'coerce': [_coerce.to_variant]}, 'respectname': {'type': 'cBool'}, 'timesort': {'type': 'cBool'}, 'copypointing': {'type': 'cBool'}, 'visweightscale': {'type': 'cFloatVec', 'coerce': [_coerce.to_list,_coerce.to_floatvec]}, 'forcesingleephemfield': {'type': 'cVariant', 'coerce': [_coerce.to_variant]}}

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 21 + 4
        output = [ ]
        addon = ''
        first_addon = True
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- return nonsubparam values ----------------------------------------------

    def __forcesingleephemfield_dflt( self, glb ):
        return ''

    def __forcesingleephemfield( self, glb ):
        if 'forcesingleephemfield' in glb: return glb['forcesingleephemfield']
        return ''

    def __vis_dflt( self, glb ):
        return ''

    def __vis( self, glb ):
        if 'vis' in glb: return glb['vis']
        return ''

    def __dirtol_dflt( self, glb ):
        return ''

    def __dirtol( self, glb ):
        if 'dirtol' in glb: return glb['dirtol']
        return ''

    def __timesort_dflt( self, glb ):
        return True

    def __timesort( self, glb ):
        if 'timesort' in glb: return glb['timesort']
        return True

    def __datacolumn_dflt( self, glb ):
        return 'corrected'

    def __datacolumn( self, glb ):
        if 'datacolumn' in glb: return glb['datacolumn']
        return 'corrected'

    def __respectname_dflt( self, glb ):
        return False

    def __respectname( self, glb ):
        if 'respectname' in glb: return glb['respectname']
        return False

    def __keep_orig_ms_dflt( self, glb ):
        return True

    def __keep_orig_ms( self, glb ):
        if 'keep_orig_ms' in glb: return glb['keep_orig_ms']
        return True

    def __visweightscale_dflt( self, glb ):
        return [  ]

    def __visweightscale( self, glb ):
        if 'visweightscale' in glb: return glb['visweightscale']
        return [  ]

    def __concatvis_dflt( self, glb ):
        return ''

    def __concatvis( self, glb ):
        if 'concatvis' in glb: return glb['concatvis']
        return ''

    def __copypointing_dflt( self, glb ):
        return True

    def __copypointing( self, glb ):
        if 'copypointing' in glb: return glb['copypointing']
        return True

    def __cols2rm_dflt( self, glb ):
        return 'model,corrected'

    def __cols2rm( self, glb ):
        if 'cols2rm' in glb: return glb['cols2rm']
        return 'model,corrected'

    def __freqtol_dflt( self, glb ):
        return ''

    def __freqtol( self, glb ):
        if 'freqtol' in glb: return glb['freqtol']
        return ''



    #--------- return inp/go default --------------------------------------------------


    #--------- return subparam values -------------------------------------------------


    #--------- subparam inp output ----------------------------------------------------
    def __vis_inp(self):
        description = ''
        value = self.__vis( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'vis': value},{'vis': self.__schema['vis']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('vis',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __concatvis_inp(self):
        description = ''
        value = self.__concatvis( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'concatvis': value},{'concatvis': self.__schema['concatvis']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('concatvis',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __datacolumn_inp(self):
        description = ''
        value = self.__datacolumn( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'datacolumn': value},{'datacolumn': self.__schema['datacolumn']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('datacolumn',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __keep_orig_ms_inp(self):
        description = ''
        value = self.__keep_orig_ms( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'keep_orig_ms': value},{'keep_orig_ms': self.__schema['keep_orig_ms']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('keep_orig_ms',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __cols2rm_inp(self):
        description = ''
        value = self.__cols2rm( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'cols2rm': value},{'cols2rm': self.__schema['cols2rm']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('cols2rm',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __freqtol_inp(self):
        description = ''
        value = self.__freqtol( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'freqtol': value},{'freqtol': self.__schema['freqtol']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('freqtol',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __dirtol_inp(self):
        description = ''
        value = self.__dirtol( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'dirtol': value},{'dirtol': self.__schema['dirtol']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('dirtol',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __respectname_inp(self):
        description = ''
        value = self.__respectname( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'respectname': value},{'respectname': self.__schema['respectname']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('respectname',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __timesort_inp(self):
        description = ''
        value = self.__timesort( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'timesort': value},{'timesort': self.__schema['timesort']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('timesort',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __copypointing_inp(self):
        description = ''
        value = self.__copypointing( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'copypointing': value},{'copypointing': self.__schema['copypointing']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('copypointing',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __visweightscale_inp(self):
        description = ''
        value = self.__visweightscale( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'visweightscale': value},{'visweightscale': self.__schema['visweightscale']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('visweightscale',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __forcesingleephemfield_inp(self):
        description = ''
        value = self.__forcesingleephemfield( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'forcesingleephemfield': value},{'forcesingleephemfield': self.__schema['forcesingleephemfield']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-21.21s = %s%-23s%s' % ('forcesingleephemfield',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))

    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'dirtol' in glb: del glb['dirtol']
        if 'concatvis' in glb: del glb['concatvis']
        if 'cols2rm' in glb: del glb['cols2rm']
        if 'datacolumn' in glb: del glb['datacolumn']
        if 'keep_orig_ms' in glb: del glb['keep_orig_ms']
        if 'freqtol' in glb: del glb['freqtol']
        if 'vis' in glb: del glb['vis']
        if 'forcesingleephemfield' in glb: del glb['forcesingleephemfield']
        if 'visweightscale' in glb: del glb['visweightscale']
        if 'timesort' in glb: del glb['timesort']
        if 'respectname' in glb: del glb['respectname']
        if 'copypointing' in glb: del glb['copypointing']


    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# concateovsa -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__vis_inp( )
        self.__concatvis_inp( )
        self.__datacolumn_inp( )
        self.__keep_orig_ms_inp( )
        self.__cols2rm_inp( )
        self.__freqtol_inp( )
        self.__dirtol_inp( )
        self.__respectname_inp( )
        self.__timesort_inp( )
        self.__copypointing_inp( )
        self.__visweightscale_inp( )
        self.__forcesingleephemfield_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,file=None):
        from casashell.private.stack_manip import find_frame
        from runpy import run_path
        filename = None
        if file is None:
            if os.path.isfile("concateovsa.last"):
                filename = "concateovsa.last"
        elif isinstance(file, str):
            if os.path.isfile(file):
                filename = file
        if filename is not None:
            glob = find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file, setting defaults instead...")
            self.set_global_defaults( )

    def __call__( self, vis=None, concatvis=None, datacolumn=None, keep_orig_ms=None, cols2rm=None, freqtol=None, dirtol=None, respectname=None, timesort=None, copypointing=None, visweightscale=None, forcesingleephemfield=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('concateovsa.pre')
        _postfile = os.path.realpath('concateovsa.last')
        _return_result_ = None
        _arguments = [vis,concatvis,datacolumn,keep_orig_ms,cols2rm,freqtol,dirtol,respectname,timesort,copypointing,visweightscale,forcesingleephemfield]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if vis is not None: local_global['vis'] = vis
            if concatvis is not None: local_global['concatvis'] = concatvis
            if datacolumn is not None: local_global['datacolumn'] = datacolumn
            if keep_orig_ms is not None: local_global['keep_orig_ms'] = keep_orig_ms
            if cols2rm is not None: local_global['cols2rm'] = cols2rm
            if freqtol is not None: local_global['freqtol'] = freqtol
            if dirtol is not None: local_global['dirtol'] = dirtol
            if respectname is not None: local_global['respectname'] = respectname
            if timesort is not None: local_global['timesort'] = timesort
            if copypointing is not None: local_global['copypointing'] = copypointing
            if visweightscale is not None: local_global['visweightscale'] = visweightscale
            if forcesingleephemfield is not None: local_global['forcesingleephemfield'] = forcesingleephemfield

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['vis'] = self.__vis( local_global )
            _invocation_parameters['concatvis'] = self.__concatvis( local_global )
            _invocation_parameters['datacolumn'] = self.__datacolumn( local_global )
            _invocation_parameters['keep_orig_ms'] = self.__keep_orig_ms( local_global )
            _invocation_parameters['cols2rm'] = self.__cols2rm( local_global )
            _invocation_parameters['freqtol'] = self.__freqtol( local_global )
            _invocation_parameters['dirtol'] = self.__dirtol( local_global )
            _invocation_parameters['respectname'] = self.__respectname( local_global )
            _invocation_parameters['timesort'] = self.__timesort( local_global )
            _invocation_parameters['copypointing'] = self.__copypointing( local_global )
            _invocation_parameters['visweightscale'] = self.__visweightscale( local_global )
            _invocation_parameters['forcesingleephemfield'] = self.__forcesingleephemfield( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default
            

        else:
            # invoke with inp/go semantics
            _invocation_parameters['vis'] = self.__vis( self.__globals_( ) )
            _invocation_parameters['concatvis'] = self.__concatvis( self.__globals_( ) )
            _invocation_parameters['datacolumn'] = self.__datacolumn( self.__globals_( ) )
            _invocation_parameters['keep_orig_ms'] = self.__keep_orig_ms( self.__globals_( ) )
            _invocation_parameters['cols2rm'] = self.__cols2rm( self.__globals_( ) )
            _invocation_parameters['freqtol'] = self.__freqtol( self.__globals_( ) )
            _invocation_parameters['dirtol'] = self.__dirtol( self.__globals_( ) )
            _invocation_parameters['respectname'] = self.__respectname( self.__globals_( ) )
            _invocation_parameters['timesort'] = self.__timesort( self.__globals_( ) )
            _invocation_parameters['copypointing'] = self.__copypointing( self.__globals_( ) )
            _invocation_parameters['visweightscale'] = self.__visweightscale( self.__globals_( ) )
            _invocation_parameters['forcesingleephemfield'] = self.__forcesingleephemfield( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-21s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#concateovsa( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _return_result_ = _concateovsa_t( _invocation_parameters['vis'],_invocation_parameters['concatvis'],_invocation_parameters['datacolumn'],_invocation_parameters['keep_orig_ms'],_invocation_parameters['cols2rm'],_invocation_parameters['freqtol'],_invocation_parameters['dirtol'],_invocation_parameters['respectname'],_invocation_parameters['timesort'],_invocation_parameters['copypointing'],_invocation_parameters['visweightscale'],_invocation_parameters['forcesingleephemfield'] )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('concateovsa')
            casalog.post("Exception Reported: Error in concateovsa: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            _return_result_ = False
        try:
            os.rename(_prefile,_postfile)
        except: pass
        return _return_result_

concateovsa = _concateovsa( )




"""
casa_compat.py

Provides a uniform interface for importing CASA (Common Astronomy Software Applications) tools across different versions and Python environments. It supports both the monolithic (CASA 4/5/6) and modular (CASA 6+) installations.

This script dynamically imports CASA components, addressing the architectural changes between versions. In monolithic installations, components are available as instantiated objects. In modular installations, components are accessed through `casatools` and `casatasks`.

Function:
- get_casa_tools(alias_list): Returns a dictionary of requested CASA tool instances. It accepts a list of tool aliases and handles dynamic import or built-in object access, ensuring compatibility across CASA versions.

Parameters:
- alias_list (list of str): Aliases for CASA tools to import. Defaults to a common set of tools.

Returns:
- dict: Mapping of tool aliases to their instances or objects.

Usage example:
    casa_tools = get_casa_tools(['tbtool', 'mstool', 'qatool', 'iatool', 'rgtool', 'msmdtool', 'smtool', 'metool'])
    for alias, instance in casa_tools.items():
        print(f"{alias}: {instance}")

The function uses `importlib` for CASA 6+ and falls back to direct access in earlier versions or interactive sessions.
"""

# Attempt dynamic imports based on the version and available modules
try:
    # Use importlib for dynamic imports in CASA 6+ environments
    import importlib
except ImportError:
    # Fallback for environments where importlib is not available (should be rare in practice)
    pass

# Define a mapping between tool aliases and their actual package names in casatools
tool_mapping = {
    'tbtool': 'table',
    'mstool': 'ms',
    'qatool': 'quanta',
    'iatool': 'image',
    'rgtool': 'regionmanager',
    'msmdtool': 'msmetadata',
    'smtool': 'simulator',
    'metool': 'measures',
    'cltool': 'componentlist',
}

def check_dependencies():
    missing_packages = []
    try:
        import aipy
    except ImportError:
        missing_packages.append("aipy-eovsa")

    try:
        import eovsapy
    except ImportError:
        missing_packages.append("eovsapy")

    if missing_packages:
        raise ImportError(
            "The following package(s) are required to use this function: {}. "
            "Please install them to proceed.".format(", ".join(missing_packages))
        )


def import_casatools(alias_list=['tbtool', 'mstool', 'qatool', 'iatool', 'rgtool', 'msmdtool', 'smtool', 'metool']):
    """
    Dynamically imports and returns CASA tools specified by their aliases.

    Parameters:
    alias_list (list of str): Aliases of the CASA tools to be imported and returned.

    Returns:
    dict: A dictionary with keys as tool aliases and values as the imported modules or objects.
    """
    tools = {}
    for alias in alias_list:
        if alias in tool_mapping:
            try:
                try:
                    # Dynamically import the module and create an instance of the tool
                    module = importlib.import_module('casatools')
                    tool_name = tool_mapping[alias]
                    tool_instance = getattr(module, tool_name)
                    tools[alias] = tool_instance
                except:
                    tools[alias] = vars()[alias]
                    # Fallback logic for older Python versions or CASA versions; adjust as needed
                    pass
            except ImportError as e:
                print(f"Error importing {tool_name} as {alias}: {e}")
                # Optionally handle the import error (e.g., skip this tool, log the error, etc.)
        else:
            print(f"No mapping found for alias: {alias}")
            # Optionally handle the case where an alias is not recognized

    return tools

def import_casatasks(*task_names):
    """
    Dynamically imports specified CASA tasks from the casatasks module. This is designed to uniformly import CASA tasks
    for both monolithic and modular CASA installations, addressing the issue where direct task import is not supported
    in monolithic CASA versions.

    Intended for modular CASA 6+ installations where tasks are accessed via casatasks. For monolithic CASA
    (versions 4/5/6), this function provides a unified interface, though direct imports are handled via the global namespace.

    Parameters:
    - task_names (str): Names of CASA tasks to import.

    Returns:
    - dict: Mapping of task names to their corresponding functions.

    Raises:
    - ImportError: If a task is not found in casatasks.
    """

    imported_tasks = {}
    try:
        from casatasks import __dict__ as casatasks_dict
        for task_name in task_names:
            if task_name in casatasks_dict:
                imported_tasks[task_name] = casatasks_dict[task_name]
        # Printing imported tasks
        if imported_tasks:
            print(f"Imported CASA tasks from casatasks: {', '.join(imported_tasks.keys())}.")
        else:
            print("No tasks were imported from casatasks.")
    except ImportError as e:
        print(f"Error importing from casatasks: {e}")
    return imported_tasks


# __name__ = "private"
# __all__ = [ 'task_ptclean6','task_subvs' ]
# from .task_ptclean6 import ptclean6
# from .task_subvs import subvs

# __name__ = "private"
# __all__ = ['task_ptclean6', 'task_subvs', 'task_importeovsa', 'task_calibeovsa', 'task_concateovsa']
#
# from .task_ptclean6 import ptclean6
# from .task_subvs import subvs
# from .task_importeovsa import importeovsa
# from .task_calibeovsa import calibeovsa
# from .task_concateovsa import concateovsa



from astropy.time import Time
from sunpy import map as smap
import numpy as np
import astropy.units as u

def readfits(eofile):
    '''
    read eovsa image fits, adjust the date-obs to the mid time.
    :param eofile:
    :return:
    '''
    eomap_ = smap.Map(eofile)
    tbeg = Time(eomap_.date)
    newmeta = eomap_.meta
    tmid = Time(tbeg.mjd + newmeta['exptime'] / 24. / 3600 / 2, format='mjd')
    newmeta['date-obs'] = tmid.isot
    eomap = smap.Map(eomap_.data, newmeta)
    return eomap

def get_all_coordinate_from_map(sunmap):
    x, y = np.meshgrid(*[np.arange(v.value) for v in sunmap.dimensions]) * u.pix
    return sunmap.pixel_to_world(x, y)


##################### generated by xml-casa (v2) from calibeovsa.xml ################
##################### f295775a00a77dd77b81d6e69924145c ##############################
from __future__ import absolute_import
import numpy
from casatools.typecheck import CasaValidator as _val_ctor
_pc = _val_ctor( )
from casatools.coercetype import coerce as _coerce
from casatools.errors import create_error_string
from .private.task_calibeovsa import calibeovsa as _calibeovsa_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log
from casatasks.private.task_logging import except_log as _except_log

class _calibeovsa:
    """
    calibeovsa ---- Calibrating EOVSA one or more measurement sets using calibration products in the SQL database.

    
    Calibrating EOVSA one or more measurement sets using calibration products in the SQL database. This task currently only works on pipeline.
    

    --------- parameter descriptions ---------------------------------------------

    vis          input EOVSA (uncalibrated) measurement set(s).
    caltype      Types of calibrations to perform
    caltbdir     Directory to place calibration tables.
    interp       Temporal interpolation for phacal table(s) (nearest or linear)
    docalib      If False, only create the calibration tables but do not perform applycal.
    doflag       If true then perform flagging.
    flagant      Antennas to be flagged. Follow CASA syntax of "antenna".
    doimage      If True, produce a quicklook image after calibration (sunpy must be installed).
    imagedir     directory to place output images. Default current directory.
    antenna      antenna/baselines to be used for imaging. Follow CASA syntax of "antenna".
    timerange    Timerange to be imaged. Follow CASA syntax of "timerange". Default is the entire duration of the ms.
    spw          spectral windows to be imaged. Follow CASA syntax of "spw".
    stokes       stokes to be imaged. Follow CASA syntax of "stokes".
    dosplit      If True, plit the corrected data column as output visibility file.
    outputvis    Name of output visibility file. Default is the name of the first vis file ended with ".corrected.ms".
    doconcat     If True, and if more than one visibility dataset provided, concatenate all into one visibility.
    concatvis    Name of output visibility file. Default is the name of the first + last vis file ended with ".corrected.ms".
    keep_orig_ms Keep the original seperated ms datasets after split?

    --------- examples -----------------------------------------------------------

    
    Calibrating EOVSA one or more measurement sets using calibration products in the SQL database.
    
    Detailed Keyword arguments:
    
    vis -- Name of input EOVSA measurement set dataset(s)
    default: none. Must be supplied
    example: vis = 'IDB20160524000518.ms'
    example: vis = ['IDB20160524000518.ms','IDB20160524000528.ms']
    
    caltype -- list. Type of calibrations to be applied.
    'refpha': reference phase calibration
    'refamp': reference amplitude calibration (not used anymore)
    'phacal': daily phase calibration
    'fluxcal': flux calibration based on total-power measurements
    default value: ['refpha','phacal']
    *** note fluxcal is already implemented in udb_corr when doing importeovsa, should not be used anymore ****
    *** pipeline only uses ['refpha','phacal']
    
    caltbdir -- string. Place to hold calibration tables. Default is current directory. Pipeline should use /data1/eovsa/caltable
    
    interp -- string. How interpolation is done for phacal? 'nearest' or 'linear'
    
    docalib -- boolean. Default True. If False, only create the calibration tables but do not perform applycal
    
    doflag -- boolean. Default True. Peforming flags?
    
    flagant -- string. Follow CASA antenna selection syntax. Default '13~15'.
    
    doimage -- boolean. Default False. If true, make a quicklook image using the specified time range and specified spw range
    
    imagedir -- string. Directory to place the output image.
    
    antenna -- string. Default '0~12'. Antenna/baselines to be used for imaging. Follow CASA antenna selection syntax.
    
    timerange -- string. Default '' (the whole duration of the visibility data). Follow CASA timerange syntax.
    e.g., '2017/07/11/20:16:00~2017/07/11/20:17:00'
    
    spw -- string. Default '1~3'. Follow CASA spw selection syntax.
    
    stokes -- string. Which stokes for the quicklook image. CASA syntax. Default 'XX'
    
    dosplit -- boolean. Split the corrected data column?
    
    outputvis -- string. Output visibility file after split
    
    doconcat -- boolean. If more than one visibility dataset provided, concatenate all into one or make separate outputs if True
    
    concatvis -- string. Output visibility file after concatenation
    
    keep_orig_ms -- boolean. Default True. Inherited from suncasa.eovsa.concateovsa.
    Keep the original seperated ms datasets after concatenation?
    
    


    """

    _info_group_ = """Calibration"""
    _info_desc_ = """Calibrating EOVSA one or more measurement sets using calibration products in the SQL database."""

    def __call__( self, vis='', caltype=[  ], caltbdir='', interp='nearest', docalib=True, doflag=True, flagant='13~15', doimage=False, imagedir='.', antenna='0~12', timerange='', spw='1~3', stokes='XX', dosplit=False, outputvis='', doconcat=False, concatvis='', keep_orig_ms=True ):
        schema = {'vis': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'caltype': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'caltbdir': {'type': 'cStr', 'coerce': _coerce.to_str}, 'interp': {'type': 'cStr', 'coerce': _coerce.to_str}, 'docalib': {'type': 'cBool'}, 'doflag': {'type': 'cBool'}, 'flagant': {'type': 'cStr', 'coerce': _coerce.to_str}, 'doimage': {'type': 'cBool'}, 'imagedir': {'type': 'cStr', 'coerce': _coerce.to_str}, 'antenna': {'type': 'cStr', 'coerce': _coerce.to_str}, 'timerange': {'type': 'cStr', 'coerce': _coerce.to_str}, 'spw': {'type': 'cStr', 'coerce': _coerce.to_str}, 'stokes': {'type': 'cStr', 'coerce': _coerce.to_str}, 'dosplit': {'type': 'cBool'}, 'outputvis': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'doconcat': {'type': 'cBool'}, 'concatvis': {'type': 'cStr', 'coerce': _coerce.to_str}, 'keep_orig_ms': {'type': 'cBool'}}
        doc = {'vis': vis, 'caltype': caltype, 'caltbdir': caltbdir, 'interp': interp, 'docalib': docalib, 'doflag': doflag, 'flagant': flagant, 'doimage': doimage, 'imagedir': imagedir, 'antenna': antenna, 'timerange': timerange, 'spw': spw, 'stokes': stokes, 'dosplit': dosplit, 'outputvis': outputvis, 'doconcat': doconcat, 'concatvis': concatvis, 'keep_orig_ms': keep_orig_ms}
        assert _pc.validate(doc,schema), create_error_string(_pc.errors)
        _logging_state_ = _start_log( 'calibeovsa', [ 'vis=' + repr(_pc.document['vis']), 'caltype=' + repr(_pc.document['caltype']), 'caltbdir=' + repr(_pc.document['caltbdir']), 'interp=' + repr(_pc.document['interp']), 'docalib=' + repr(_pc.document['docalib']), 'doflag=' + repr(_pc.document['doflag']), 'flagant=' + repr(_pc.document['flagant']), 'doimage=' + repr(_pc.document['doimage']), 'imagedir=' + repr(_pc.document['imagedir']), 'antenna=' + repr(_pc.document['antenna']), 'timerange=' + repr(_pc.document['timerange']), 'spw=' + repr(_pc.document['spw']), 'stokes=' + repr(_pc.document['stokes']), 'dosplit=' + repr(_pc.document['dosplit']), 'outputvis=' + repr(_pc.document['outputvis']), 'doconcat=' + repr(_pc.document['doconcat']), 'concatvis=' + repr(_pc.document['concatvis']), 'keep_orig_ms=' + repr(_pc.document['keep_orig_ms']) ] )
        task_result = None
        try:
            task_result = _calibeovsa_t( _pc.document['vis'], _pc.document['caltype'], _pc.document['caltbdir'], _pc.document['interp'], _pc.document['docalib'], _pc.document['doflag'], _pc.document['flagant'], _pc.document['doimage'], _pc.document['imagedir'], _pc.document['antenna'], _pc.document['timerange'], _pc.document['spw'], _pc.document['stokes'], _pc.document['dosplit'], _pc.document['outputvis'], _pc.document['doconcat'], _pc.document['concatvis'], _pc.document['keep_orig_ms'] )
        except Exception as exc:
            _except_log('calibeovsa', exc)
            raise
        finally:
            task_result = _end_log( _logging_state_, 'calibeovsa', task_result )
        return task_result

calibeovsa = _calibeovsa( )




def lightcurves(timerange, outdir='./', specfile=None, goes=True, hessifile=None, fermifile=None, ylog=False, hessi_smoth=0, dspec_cmap='cubehelix',
                vmax=None, vmin=None):
    from sunpy.lightcurve import GOESLightCurve
    from sunpy.time import TimeRange, parse_time
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from astropy.time import Time
    import numpy as np
    import numpy.ma as ma
    from scipy.signal import medfilt
    import matplotlib.colors as colors
    from scipy.io import readsav
    from suncasa.utils import DButil
    import os

    timerange = Time(timerange)
    if hessifile:
        if not os.path.exists(hessifile):
            hessi_script = 'HESSI_lc.pro'
            print('Run the script {} in SSWIDL to download RHESSI summary file first!'.format(hessi_script))
            fi = open(hessi_script, 'wb')
            fi.write(b'search_network, /enable \n')
            text = "time_range = ['{}','{}'] \n".format(timerange[0].datetime.strftime('%Y-%b-%d %H:%M:%S'),
                                                          timerange[1].datetime.strftime('%Y-%b-%d %H:%M:%S'))

            fi.write(text.encode())
            fi.write(b'obs_obj = hsi_obs_summary(obs_time_interval=time_range) \n')
            fi.write(b'data = obs_obj->getdata() \n')
            fi.write(b'info = obs_obj->get( / info) \n')
            fi.write(b'obs_data = data.countrate \n')
            fi.write(b'obs_times = obs_obj->getdata(/time) \n')
            fi.write(b'obs_times_str  = anytim(obs_times,/CCSDS) \n')
            fi.write(b'obs_energies  = fltarr(2,n_elements(info.energy_edges)-1) \n')
            fi.write(b'for ll=0,n_elements(info.energy_edges)-2 do begin \n')
            fi.write(b'    obs_energies[0,ll]=info.energy_edges[ll] \n')
            fi.write(b'    obs_energies[1,ll]=info.energy_edges[ll+1] \n')
            fi.write(b'endfor \n')
            text = 'save,filename="{}",OBS_DATA,OBS_ENERGIES,obs_times_str \n'.format(hessifile)
            fi.write(text.encode())
            fi.write(b'end \n')
            fi.close()
            return -1
        hessi = readsav(hessifile)
        hessi_tim = Time(list(hessi['obs_times_str']))
        hessi_tim_plt = hessi_tim.plot_date

    specdata = np.load(specfile)
    spec = specdata['spec']
    if len(spec.shape) == 4:
        (npol, nbl, nfreq, ntim) = spec.shape
    else:
        (nfreq, ntim) = spec.shape
    spec = np.mean(np.mean(spec, axis=0), axis=0)
    freq = specdata['freq']
    freqghz = freq / 1e9
    spec_tim = Time(specdata['tim'] / 3600. / 24., format='mjd')

    fidx_plt = np.linspace(0, nfreq - nfreq / 8.0 / 2.0, 8).astype(int) + nfreq / 8.0 / 2.0

    try:
        plt.style.use('seaborn-bright')
        params = {'font.size': 8, 'axes.grid': False, 'axes.facecolor': 'w', 'xtick.color': '#555555', 'ytick.color': '#555555',
                  'xtick.major.size': 2.0, 'ytick.major.size': 2.0, 'xtick.minor.size': 1.0, 'ytick.minor.size': 1.0, 'axes.axisbelow': False,
                  'axes.xmargin': 0.0, 'axes.ymargin': 0.0, 'axes.linewidth': 0.5, 'xtick.major.pad': 1.5, 'ytick.major.pad': 1.5,
                  'lines.linewidth': 1.0}
        mpl.rcParams.update(params)
    except:
        pass

    if goes:
        tr = TimeRange(timerange.iso)
        goes = GOESLightCurve.create(tr)
        dates = mpl.dates.date2num(parse_time(goes.data.index))

    tr_plt = Time(timerange)

    fig, axs = plt.subplots(nrows=3, ncols=1, sharex=True, figsize=(8, 6))
    ax = axs[0]
    tidx_hessi, = np.where((hessi_tim >= tr_plt[0]) & (hessi_tim <= tr_plt[1]))
    for idx, eg in enumerate(hessi['obs_energies']):
        flux_hessi = ma.masked_array(hessi['obs_data'][tidx_hessi, idx])
        if hessi_smoth > 0:
            flux_hessi = DButil.smooth(flux_hessi, 20)
            flux_hessi = flux_hessi / np.nanmax(flux_hessi)
            ax.step(hessi_tim_plt[tidx_hessi], DButil.smooth(flux_hessi, hessi_smoth), label='{:.0f}-{:.0f} keV'.format(eg[0], eg[1]))
        else:
            ax.step(hessi_tim_plt[tidx_hessi], flux_hessi, label='{:.0f}-{:.0f} keV'.format(eg[0], eg[1]))

    if ylog:
        ax.set_yscale("log", nonposy='clip')
    else:
        ax.set_yscale("linear", nonposy='clip')
    ax.legend()
    ax.set_ylabel('Count Rate [s$^{-1}$ detector$^{-1}$ ]')

    ax = axs[1]
    tidx_goes, = np.where((dates >= tr_plt[0].plot_date) & (dates <= tr_plt[1].plot_date))
    if goes:
        ax.plot(dates[tidx_goes], goes.data['xrsb'][tidx_goes] / np.nanmax(goes.data['xrsb'][tidx_goes]), label='GOES 1.0--8.0 $\AA$')
        ax.plot(dates[tidx_goes], goes.data['xrsa'][tidx_goes] / np.nanmax(goes.data['xrsa'][tidx_goes]), label='GOES 0.5--4.0 $\AA$')
    tidx_spec, = np.where((spec_tim >= tr_plt[0]) & (spec_tim <= tr_plt[1]))
    if len(tidx_spec) > 1:
        spec_tim_plt = spec_tim[tidx_spec[0]:tidx_spec[-1]].plot_date
        flux_colors = []
        for idx, fidx in enumerate(np.round(fidx_plt).astype(int)):
            flux_plt = medfilt(spec[min(fidx, nfreq - 1), tidx_spec[0]:tidx_spec[-1]], 7)
            p = ax.plot(spec_tim_plt, flux_plt / np.nanmax(flux_plt), label='{:.2f} GHz'.format(freqghz[min(fidx, nfreq - 1)]))
            flux_colors.append(p[0].get_color())
        ax.set_ylabel('Flux (Normalized)')
        ax.set_ylim(0, 1.1)
        ax.legend()

        ax = axs[2]
        spec_plt = spec[:, tidx_spec[0]:tidx_spec[-1]]
        ax.pcolormesh(spec_tim_plt, freqghz, spec_plt, cmap=dspec_cmap, norm=colors.LogNorm(vmax=vmax, vmin=vmin))
        ax.set_ylabel('Frequency [GHz]')
        formatter = mpl.dates.DateFormatter('%H:%M:%S')
        ax.xaxis.set_major_formatter(formatter)
        ax.fmt_xdata = formatter
        ax.set_xlim(tr_plt.plot_date)
        ax.set_xlabel('Start time ({})'.format(tr_plt[0].datetime.strftime('%d-%b-%y %H:%M:%S')))
        for idx, fidx in enumerate(np.round(fidx_plt).astype(int)):
            ax.axhline(freqghz[min(fidx, nfreq - 1)], color=flux_colors[idx], ls=':')
    else:
        print('Warning: No radio data in the timerange. Proceed without dynamic spectrum.')

    fig.tight_layout()
    fig.subplots_adjust(hspace=0.06)
    imgdir = outdir + '/fig01_{}-{}.png'.format(tr_plt[0].datetime.strftime('%H%M%S'), tr_plt[1].datetime.strftime('%H%M%S'))
    fig.savefig(imgdir, dpi=200)
    print('Save image to ' + imgdir)



import numpy as np
from tqdm import tqdm
import os
from shutil import rmtree

from ..casa_compat import import_casatools,import_casatasks

tasks = import_casatasks('split', 'tclean', 'casalog', 'clearcal', 'gaincal')
split = tasks.get('split')
tclean = tasks.get('tclean')
casalog = tasks.get('casalog')
clearcal = tasks.get('clearcal')
gaincal = tasks.get('gaincal')

tools = import_casatools(['tbtool', 'mstool', 'qatool'])
tbtool = tools['tbtool']
mstool = tools['mstool']
qatool = tools['qatool']
tb = tbtool()
ms = mstool()
qa = qatool()

def get_bandinfo(msfile, spw=None, returnbdinfo=False):
    '''
    get center frequencies of all spectral windows for msfile
    spw: [option] return the cfreq of spw. spw can be a a string or a list of string.
    The syntax of spw follows the standard spw Parameter in CASA
    if spw is not provided, return the cfreq of all spws in the msfile.
    return cfreqs is in GHz
    if returnbounds is True, return a dictionary including comprehensive freq information of the ms.
    '''

    ms.open(msfile)
    spwInfo = ms.getspectralwindowinfo()
    nspw = len(spwInfo.keys())
    reffreqs = []
    bdwds = []
    chanwds = []
    nchans = []
    for s in range(nspw):
        s_ = str(s)
        reffreqs.append(spwInfo[s_]['RefFreq'])
        bdwds.append(spwInfo[s_]['TotalWidth'])
        chanwds.append(spwInfo[s_]['ChanWidth'])
        nchans.append(spwInfo[s_]['NumChan'])
    reffreqs = np.array(reffreqs) / 1e9
    bdwds = np.array(bdwds) / 1e9
    chanwds = np.array(chanwds) / 1e9
    nchans = np.array(nchans)
    cfreqs = reffreqs + bdwds / 2.0 - chanwds / 2.0
    bdinfo = {'bounds_all': np.hstack((reffreqs, reffreqs[-1] + bdwds[-1])), 'cfreqs_all': cfreqs,
              'bounds_all_lo': reffreqs, 'bounds_all_hi': reffreqs + bdwds, 'nchans': nchans}
    if spw:
        freqbounds_lo_spw = []
        freqbounds_hi_spw = []
        cfreqs_spw = []
        for sp in spw:
            print(f'Parsing spw {sp}...')
            try:
                staql = {'spw': sp}
                ms.selectinit(reset=True)
                ms.msselect(staql, onlyparse=True)
                ndx = ms.msselectedindices()
                chan_sel = ndx['channel']
                bspw = chan_sel[0, 0]
                bchan = chan_sel[0, 1]
                espw = chan_sel[-1, 0]
                echan = chan_sel[-1, 2]
                bfreq = (spwInfo[str(bspw)]['Chan1Freq'] + spwInfo[str(bspw)]['ChanWidth'] * bchan) / 1e9
                efreq = (spwInfo[str(espw)]['Chan1Freq'] + spwInfo[str(espw)]['ChanWidth'] * echan) / 1e9
                cfreq = (bfreq + efreq) / 2.
            except ValueError:
                print("Parsing spw {} failed. Aborting...".format(sp))
                continue
            freqbounds_lo_spw.append(bfreq)
            freqbounds_hi_spw.append(efreq)
            cfreqs_spw.append(cfreq)
        cfreqs = np.array(cfreqs_spw)
        freqbounds_lo = np.array(freqbounds_lo_spw)
        freqbounds_hi = np.array(freqbounds_hi_spw)
        bdinfo['bounds_lo'] = freqbounds_lo
        bdinfo['bounds_hi'] = freqbounds_hi
        bdinfo['cfreqs'] = cfreqs

    ms.done()
    if returnbdinfo:
        return bdinfo
    else:
        return cfreqs


def get_bmsize(cfreq, refbmsize=70.0, reffreq=1.0, minbmsize=4.0):
    '''
    get beamsize at frequencies definded by cfreq based on refbmsize at reffreq
    cfreq: input frequencies at GHz
    refbmsize: reference beam size in arcsec
    reffreq: reference frequency in GHz
    minbmsize: minimum beam size in arcsec
    '''
    bmsize = refbmsize * reffreq / cfreq
    bmsize[bmsize < minbmsize] = minbmsize
    return bmsize


def get_trange(msfile):
    from astropy.time import Time
    tb.open(msfile)
    tr = np.array([tb.getcell('TIME', 0), tb.getcell('TIME', tb.nrows() - 1)]) / 24. / 3600.
    tb.close()
    return Time(tr, format='mjd')


def time2filename(msfile, timerange='', spw='', desc=False):
    from astropy.time import Time
    tb.open(msfile)
    starttim = Time(tb.getcell('TIME', 0) / 24. / 3600., format='mjd')
    endtim = Time(tb.getcell('TIME', tb.nrows() - 1) / 24. / 3600., format='mjd')
    tb.close()
    datstr = starttim.iso[:10]
    ms.open(msfile)
    metadata = ms.metadata()
    observatory = metadata.observatorynames()[0]
    ms.done()
    if timerange is None or timerange == '':
        starttim1 = starttim
        endtim1 = endtim
    else:
        (tstart, tend) = timerange.split('~')
        if tstart[2] == ':':
            starttim1 = Time(datstr + 'T' + tstart)
            endtim1 = Time(datstr + 'T' + tend)
        else:
            starttim1 = Time(qa.quantity(tstart, 'd')['value'], format='mjd')
            endtim1 = Time(qa.quantity(tend, 'd')['value'], format='mjd')
    if desc:
        tstr = starttim1.to_datetime().strftime(
            '{}_%Y%m%dT%H%M%S.%f'.format(observatory)) + endtim1.to_datetime().strftime('-%H%M%S.%f')
    else:
        midtime = Time((starttim1.mjd + endtim1.mjd) / 2., format='mjd')
        tstr = midtime.to_datetime().strftime('{}_%Y%m%dT%H%M%S.%f'.format(observatory))

    if spw:
        spstr = 'spw{}'.format(spw.replace('~', '-'))
        filename = '.'.join([tstr, spstr])
    else:
        filename = tstr

    return filename


def msclearhistory(msfile):
    '''Clears history in the a measurement sets file

    :param msfile: string
            The name of a measurement sets file

    :return:
    '''

    tb.open(msfile + '/HISTORY', nomodify=False)
    nrows = tb.nrows()
    if nrows > 0:
        tb.removerows(range(nrows))
    tb.close()


def clearflagrow(msfile, mode='clear'):
    '''

    :param msfile:
    :param mode: FLAG_ROW operation
    default: 'clear': (default) clear the FLAG_ROW
             'list': to list existing FLAG_ROW
    :return:
    '''

    if mode == 'list':
        tb.open(msfile, nomodify=True)
        a = tb.getcol('FLAG_ROW')
        nfrows = np.sum(a)
        nrows = float(len(a))
        print('{:0d} out of {:.0f} ({:.0f}%) rows are flagged in {}'.format(nfrows, nrows, nfrows / nrows * 100,
                                                                            os.path.basename(msfile)))
    elif mode == 'clear':
        tb.open(msfile, nomodify=False)
        a = tb.getcol('FLAG_ROW')
        a[:] = False
        tb.putcol('FLAG_ROW', a)
        print('reset successfully')
    tb.close()


def splitX(vis, datacolumn2='MODEL_DATA', **kwargs):
    import os
    """
    Splits specific data columns from a CASA measurement set (MS) into a new MS file,
    overcoming the limitation of splitting multiple data columns directly with CASA's standard split function.

    A temporary MS is used during processing to ensure the additional column (the `datacolumn2`) has the same shape as
    the data/corrected column in the output MS. This temporary file is removed at the end.

    :param str vis: Path to the original measurement set.
    :param str datacolumn2: The name of the additional data column to be included in the split. Defaults to 'MODEL_DATA'.
    :param kwargs: Additional keyword arguments to be passed to the CASA `split` task.
    :return: Path to the output measurement set with the specified columns split.
    :rtype: str
    """

    datacolumn2 = datacolumn2.upper()
    outmsfile = kwargs.get('outputvis', '').rstrip('/')

    # Clean up the output directory and associated flag versions
    if os.path.exists(outmsfile):
        rmtree(outmsfile)
    flagversions_path = f"{outmsfile}.flagversions"
    if os.path.exists(flagversions_path):
        rmtree(flagversions_path)

    # Perform the initial split to create the output MS with the data/corrected column into the data column
    split(vis=vis, **kwargs)

    # Prepare for the second split
    tmpms_file = f'{vis}.tmpms'
    kwargs2 = {k: v for k, v in kwargs.items() if k not in ['datacolumn', 'outputvis']}
    kwargs2.update({'outputvis': tmpms_file, 'datacolumn': datacolumn2.replace('_DATA', '')})

    # Ensure temporary MS is removed if it exists

    if os.path.exists(tmpms_file):
        rmtree(tmpms_file)

    # Perform the second split to temporary MS
    split(vis=vis, **kwargs2)

    # Process the additional column from the temporary MS
    tb.open(tmpms_file)
    nrows = tb.nrows()
    data = []
    for row in tqdm(range(nrows), desc=f'Extracting {datacolumn2} column', ascii=True):
        data.append(tb.getcell('DATA', row))
    tb.close()

    # Add additional data column to the output MS
    clearcal(outmsfile, addmodel=True)
    tb.open(outmsfile, nomodify=False)
    for row in tqdm(range(nrows), desc=f'writing {datacolumn2} column', ascii=True):
        tb.putcell(datacolumn2, row, data[row])
    tb.close()

    # Cleanup
    rmtree(tmpms_file)
    return outmsfile


def flagcaltboutliers(caltable, limit=[]):
    import numpy as np
    import numpy.ma as ma
    # def removeOutliers(x, outlierConstant):
    #     a = np.array(x)
    #     idx, = np.where(np.diff(np.sort(datamag[0, 0, :]))>)
    #     upper_quartile = np.percentile(a, 80)
    #     lower_quartile = np.percentile(a, 20)
    #     IQR = (upper_quartile - lower_quartile) * outlierConstant
    #     quartileSet = (lower_quartile - IQR, upper_quartile + IQR)
    #     return ma.masked_outside(x, quartileSet[1], quartileSet[0])

    if not os.path.exists(caltable): return 0
    if isinstance(limit, list):
        if len(limit) == 2:
            tb.open(caltable, nomodify=False)
            # subt = tb.query("ANTENNA1==1 && SPECTRAL_WINDOW_ID=10")
            # data = subt.getcol('CPARAM')
            # flag = subt.getcol('FLAG')
            # spw = subt.getcol('SPECTRAL_WINDOW_ID')
            # datamag = np.abs(data)
            # mdatamag = ma.masked_outside(datamag, limit[0], limit[1])
            # mask = np.logical_or(mdatamag.mask, flag)
            # dataidx1 = datamag<limit[0]
            # dataidx2 = datamag>limit[1]
            # mdatamag = ma.masked_array(mdatamag, mask)
            # mdatamag[0, 0, :] = removeOutliers(mdatamag[0, 0, :], 5)
            # mdatamag[1, 0, :] = removeOutliers(mdatamag[1, 0, :], 5)
            data = tb.getcol('CPARAM')
            flag = tb.getcol('FLAG')
            datamag = np.abs(data)
            dataidx1 = datamag < limit[0]
            dataidx2 = datamag > limit[1]
            flag[dataidx1] = True
            flag[dataidx2] = True
            tb.putcol('FLAG', flag)
            return 1
        else:
            print('limit must have two elements. Aborted!')
            return 0
    else:
        print('limit must be a list. Aborted!')


def modeltransfer(msfile, spw='', reference='XX', transfer='YY'):
    pol_dict = {'XX': 0, 'YY': 1, 'XY': 2, 'YX': 3}
    refidx = pol_dict[reference]
    trfidx = pol_dict[transfer]
    datams = mstool()
    datams.open(msfile, nomodify=False)

    if '~' in spw:
        sp0, sp1 = spw.split('~')
        for sp in range(int(sp0), int(sp1) + 1):
            staql = {'spw': str(sp)}
            datams.selectinit(reset=True)
            datams.msselect(staql)
            modeldata = datams.getdata(['model_data'])
            modeldata['model_data'][trfidx, ...] = modeldata['model_data'][refidx, ...]
            datams.putdata(modeldata)
        datams.close()
    else:
        datams.selectinit(reset=True)
        staql = {'spw': spw}
        datams.msselect(staql)
        modeldata = datams.getdata(['model_data'])
        modeldata['model_data'][trfidx, ...] = modeldata['model_data'][refidx, ...]
        datams.putdata(modeldata)
        datams.close()


def concat_slftb(tb_in=[], tb_out=None):
    if not tb_in:
        print('tb_in not provided. Abort...')
    if os.path.exists(tb_out):
        os.system('rm -r {}'.format(tb_out))
    os.system('cp -r {} {}'.format(tb_in[0], tb_out))
    tbdata = {}
    tb.open(tb_out)
    cols = tb.colnames()
    tb.close()
    cols.remove('WEIGHT')
    for col in cols:
        tbdata[col] = []
    for tbidx, ctb in enumerate(tb_in):
        tb.open(ctb, nomodify=True)
        tim0 = tb.getcol(cols[0])
        if len(tim0) == 0:
            continue
        else:
            for col in cols:
                if tbidx == 1 and col in ['CPARAM', 'PARAMERR', 'FLAG', 'SNR']:
                    tbdata[col].append(tb.getcol(col)[::-1, ...])
                else:
                    tbdata[col].append(tb.getcol(col))
        tb.close()

    if len(tbdata[cols[0]]) == 0:
        print('tables have no data. Return')
        return -1
    else:
        for col in cols:
            if col in ['CPARAM', 'PARAMERR', 'FLAG', 'SNR']:
                tbdata[col] = np.concatenate(tbdata[col], axis=2)
            else:
                tbdata[col] = np.concatenate(tbdata[col])
        tb.open(tb_out, nomodify=False)
        nrows = tb.nrows()
        nrows_new = len(tbdata[cols[0]])
        tb.addrows(nrows_new - nrows)
        for col in cols:
            tb.putcol(col, tbdata[col])
        tb.close()
        return tb_out


def gaincalXY(vis=None, caltable=None, pols='XXYY', msfileXY=None, gaintableXY=None, **kwargs):
    if pols == 'XXYY':
        pols = 'XX,YY'
    pols_ = pols.split(',')
    rm_msfileXY = False
    if msfileXY is None:
        rm_msfileXY = True
        msfileXY = {}
        for pol in pols_:
            msfileXY[pol] = '.'.join([vis, pol])
            if os.path.exists(msfileXY[pol]):
                os.system('rm -rf {}'.format(msfileXY[pol]))
            splitX(vis=vis, outputvis=msfileXY[pol], correlation=pol, datacolumn='data', datacolumn2='MODEL_DATA')
    if gaintableXY is not None:
        if 'gaintable' in kwargs.keys():
            kwargs.pop('gaintable')
    caltbXY = []
    for pol in pols_:
        caltb_ = '.'.join([caltable, pol])
        if gaintableXY is not None:
            kwargs['gaintable'] = gaintableXY[pol]
        gaincal(vis=msfileXY[pol], caltable=caltb_, **kwargs)
        caltbXY.append(caltb_)
    concat_slftb(caltbXY, caltable)
    if rm_msfileXY:
        for k, v in msfileXY.iteritems():
            os.system('rm -rf {}'.format(v))
    return


def getmodel(vis, spw=None):
    tb.open(vis, nomodify=True)
    subt = tb.query("DATA_DESC_ID==" + str(spw))
    model_d = subt.getcol('MODEL_DATA')
    subt.done()
    tb.done()
    return model_d


def putmodel(vis, spw=None, model=None):
    tb.open(vis, nomodify=False)
    subt = tb.query("DATA_DESC_ID==" + str(spw))
    model_d = subt.putcol('MODEL_DATA', model)
    subt.done()
    tb.done()
    return model_d



import os,sys
import numpy as np
import pickle
import warnings
from suncasa.dspec import dspec
from suncasa.utils import qlookplot
from suncasa.utils import helioimage2fits as hf
from suncasa.utils import fitsutils
import sunpy
from sunpy import map as smap
from matplotlib import gridspec as gridspec
from matplotlib import pyplot as plt
import math
import astropy.units as u
from astropy.coordinates import SkyCoord

warnings.filterwarnings("ignore")

'''
Example script for self-calibrating EOVSA flare data. 
- Input of the script is a CASA visibility measurement set which has prior flux and gain calibration
    The example can be downloaded from the following link:
        https://drive.google.com/file/d/1i46fvsCS4B7DfoTwwL1JQkBiB5MTVVVE/view?usp=share_link
- Output of the script is a self-calibrated CASA visibility measurement set, self-calibration tables, and intermediate
    products such as images after each self-calibration rounds 
- The script involves interactive mask drawing with CASA's imview, and hence need to be run under 
    a **monolithic** CASA 6.5 environment (both Python 3.6 and Python 3.8 are tested). 
    A modular CASA can be used if interactive mask drawing is not required (e.g., you already have the masks)
- Also needed is a Python 3.6/3.8 installation of suncasa, astropy, and sunpy, which can be side loaded by 
    including your Python search path, which can be found by doing the following under your Python environment:
    > import sys
    > sys.path()
  In Anaconda Python, the search path to the installed packages is usually something like
    ~/anaconda3/envs/suncasa/lib/python3.6/site-packages/
- In CASA, append the search path as follows
    > import sys
    > sys.path.append("YOUR_PYTHON_3_SEARCH_PATH")
- Run the script in CASA
    > execfile('slfcal_flare_example.py')

History:
    2019-May-16 Bin Chen (bin.chen@njit.edu)
       Initial creation of an example script as part of the EOVSA tutorial at
       the RHESSI XVIII Workshop (http://rhessi18.umn.edu/). 
       The original script used CASA 5.4 with Python 2.7. Link to the script:
            https://github.com/binchensun/eovsa-tutorial/blob/master/rhessi18/slfcal_example.py
       
    2023-Jan-08 Bin Chen (bin.chen@njit.edu)
       Major updates for compatibility with CASA 6.5 / Python 3, along with many feature enhancements.
       This script is used for the flare self-calibration tutorial at the 2023 EOVSA/GX/FASR Workshop. 
       Link to this script on Github:
            https://github.com/binchensun/suncasa/blob/master/examples/eovsa_flare_slfcal_example.py
                         
'''


# ################ The following are function blocks needed to run the script. Change at your own risk. ###########
###################################################################################################################
def draw_masks(slfcalms, spwrans_mask=None, maskdir=None, maskfile='masks.p',
               restoringbeam=['6arcsec'], timerange=None, phasecenter='', npix=512, gain=0.05, uvrange=''):
    """Function for drawing masks interactively"""
    antennas = ''
    pol = 'I'
    imgprefix = maskdir + 'slf_t0'

    # step 1: set up the clean masks
    img_init = imgprefix + '_mask_'
    os.system('rm -rf ' + img_init + '*')

    masks = []
    imnames = []
    for spwran in spwrans_mask:
        imname = img_init + spwran.replace('~', '-')
        try:
            tclean(vis=slfcalms, antenna=antennas, imagename=imname, spw=spwran,
                   uvrange=uvrange, specmode='mfs', timerange=timerange, imsize=[npix],
                   cell=['1arcsec'], niter=1000, gain=gain, stokes=stokes,
                   restoringbeam=restoringbeam, phasecenter=phasecenter, weighting='briggs', robust=1.0,
                   interactive=True, datacolumn='corrected', pbcor=True)
            imnames.append(imname + '.image')
            masks.append(imname + '.mask')
            clnjunks = ['.flux', '.model', '.psf', '.residual']
            for clnjunk in clnjunks:
                if os.path.exists(imname + clnjunk):
                    os.system('rm -rf ' + imname + clnjunk)
        except:
            print('error in cleaning spw: ' + spwran)

    pickle.dump(masks, open(maskfile, 'wb'))
    return maskfile


def make_images(slfcalms, msinfo=None, spws=None, spwrans_mask=None, usemask=True, maskfile=None,
                imgprefix='./slf', round=0, trange='', uvrange='', antennas='',
                npix=512, cell=['1arcsec'], niter=1000, gain=0.05, robust=1.0, sbeam_1GHz=40.,
                sclfactor=1, xycen=None, fov=[256, 256], stokes='I', nrow=3,
                usemsphacenter=True, observatory='EOVSA'):
    """Function for making images based on CASA tclean and registration with suncasa"""
    if not msinfo:
        msinfo = hf.read_msinfo(slfcalms)

    phasecenter, midtime = hf.calc_phasecenter_from_solxy(slfcalms, timerange=trange, xycen=xycen,
                                                          usemsphacenter=usemsphacenter, observatory=observatory)
    msmd.open(slfcalms)
    nspw = msmd.nspw()
    cfreqsghz = [msmd.meanfreq(s, unit='GHz') for s in range(nspw)]
    msmd.done()
    # Make new images after each round
    nspw = len(spws)
    ncol = math.ceil(nspw / nrow)
    xran = [xycen[0] - fov[0] / 2., xycen[0] + fov[0] / 2.]
    yran = [xycen[1] - fov[1] / 2., xycen[1] + fov[1] / 2.]
    if xran or yran:
        xyasp = (xran[1] - xran[0]) / (yran[1] - yran[0])
    ysize = 8.
    fig = plt.figure(figsize=(ncol / nrow * ysize * xyasp, ysize))
    gs = gridspec.GridSpec(nrow, ncol)
    fitsfiles = []
    for s, sp in enumerate(spws):
        print('cleaning spw: ' + sp)
        cfreqghz = cfreqsghz[int(sp)]
        # setting restoring beam size (not very useful for selfcal anyway, but just to see the results)
        bm = max(sbeam_1GHz / cfreqghz, 3.)
        print('Beam size {0:.1f}" at ref freq {1:.2f} GHz'.format(bm, cfreqghz))
        slfcal_img = imgprefix + '.spw' + sp.zfill(2) + '.slfcal' + str(round)
        # only the first round uses nearby spws for getting initial model
        spwran = sp
        if usemask:
            try:
                masks = pickle.load(open(maskfile, 'rb'))
            except:
                print('Mask file {} does not exist. Do not use any mask'.format(maskfile))
                mask = ''
            if len(masks) > 0:
                foundmask = False
                for m, spwran_mask in enumerate(spwrans_mask):
                    sbg, sed = spwran_mask.split('~')
                    if (int(sp) >= int(sbg)) and (int(sp) <= int(sed)):
                        mask = masks[m]
                        print('using mask {0:s}'.format(os.path.basename(mask)))
                        foundmask = True
                if not foundmask:
                    print('mask not defined for {}. Do use any masks.'.format(sp))
                    mask = ''
            else:
                mask = ''
        else:
            mask = ''
        try:
            tclean(vis=slfcalms, antenna=antennas, imagename=slfcal_img, uvrange=uvrange, spw=spwran,
                   specmode='mfs', timerange='', imsize=[npix], cell=cell, niter=niter,
                   gain=gain, stokes=stokes, weighting='briggs', robust=robust, phasecenter=phasecenter, mask=mask,
                   restoringbeam=[str(bm) + 'arcsec'], pbcor=False, interactive=False)
            if os.path.exists(slfcal_img + '.image'):
                fitsfile = slfcal_img + '.fits'
                hf.imreg(vis=slfcalms, msinfo=msinfo, imagefile=slfcal_img + '.image', fitsfile=fitsfile,
                         timerange=trange, usephacenter=True, geocentric=True, sclfactor=sclfactor, toTb=True,
                         verbose=False,
                         overwrite=True)
            clnjunks = ['.mask', '.flux', '.model', '.psf', '.residual', '.image', '.pb', '.image.pbcor', '.sumwt']
            for clnjunk in clnjunks:
                if os.path.exists(slfcal_img + clnjunk):
                    os.system('rm -rf ' + slfcal_img + clnjunk)
            eomap = smap.Map(fitsfile)
            ax = fig.add_subplot(gs[s], projection=eomap)
            top_right = SkyCoord(xran[1] * u.arcsec, yran[1] * u.arcsec, frame=eomap.coordinate_frame)
            bottom_left = SkyCoord(xran[0] * u.arcsec, yran[0] * u.arcsec, frame=eomap.coordinate_frame)
            eomap_sub = eomap.submap(bottom_left, top_right=top_right)
            eomap_sub.plot_settings['cmap'] = plt.get_cmap('jet')
            eomap_sub.plot(axes=ax)
            lon = ax.coords[0]
            lat = ax.coords[1]
            ax.set_title(' ')
            ax.text(0.96, 0.96, 'SPW {0:s} at {1:.1f} GHz'.format(sp, cfreqghz), transform=ax.transAxes,
                    ha='right', va='top', color='w', fontweight='bold')
            plt.subplots_adjust(left=0.1, bottom=0.08, right=0.96, top=0.94, wspace=0, hspace=0)
            if s / ncol < nrow - 1:
                ax.set_xlabel(' ')
                lon.set_ticklabel_visible(False)
            else:
                ax.set_xlabel('Solar X (arcsec)')
            if s % ncol != 0:
                ax.set_ylabel(' ')
                lat.set_ticklabel_visible(False)
            else:
                ax.set_ylabel('Solar Y (arcsec)')
            fitsfiles.append(fitsfile)
        except:
            print('error in cleaning spw: ' + sp)
            continue

    figname = imgprefix + '_t0_n{:d}.png'.format(round)
    plt.subplots_adjust(left=0.1, bottom=0.08, right=0.96, top=0.94, wspace=0, hspace=0)
    plt.suptitle("Images after round {0:d}".format(round))
    plt.savefig(figname)
    plt.close()
    return figname, fitsfiles


def plt_slftable(slfcalms, slftbs, caltypes, docombine=False, markersize=2.0):
    """
    Function for plotting self-calibration tables

    :param slfcalms: reference ms file for the spw and antenna information
    :param slftbs: a list of self-calibration gain tables
    :param caltypes: a list of self-calibration types ('a', 'p', or 'ap')
    :param docombine: if True, combine all the calibration tables to make a combined phase plot
    :param markersize: size of the plot symbols
    :return:
        1: succeeded. -1: failed
    """
    if type(slftbs) == str:
        slftbs = [slftbs]
    if type(caltypes) == str:
        caltypes = [caltypes]

    if len(slftbs) != len(caltypes):
        print('The length of slftbs must equal to caltypes. Abort...')
        return -1

    msmd.open(slfcalms)
    nspw_ms = msmd.nspw()
    cfreqsghz = [msmd.meanfreq(s, unit='GHz') for s in range(nspw_ms)]
    nant_ms = msmd.nantennas()
    msmd.done()

    nslftb = len(slftbs)
    # now read the self-calibration tables
    gains = []
    mgains = []
    flags = []
    errs = []
    snrs = []

    nrow = 4
    for n, slftb in enumerate(slftbs):
        caltype = caltypes[n]
        tb.open(slftb)
        spws = np.unique(tb.getcol('SPECTRAL_WINDOW_ID'))
        nspw = len(spws)
        ants = np.unique(tb.getcol('ANTENNA1'))
        nant = len(ants)
        ncol = math.ceil(nant / nrow)
        cparm = tb.getcol('CPARAM')
        npol, nval, ndata = cparm.shape
        gain = cparm.reshape(npol, nspw, nant)
        flag = tb.getcol('FLAG').reshape(npol, nspw, nant)
        mgain = np.ma.masked_array(gain, mask=flag)
        err = tb.getcol('PARAMERR').reshape(npol, nspw, nant)
        snr = tb.getcol('SNR').reshape(npol, nspw, nant)
        tb.close()
        if n == 0:
            gain_tot = np.copy(gain)
            flag_tot = np.copy(flag)
        else:
            gain_tot *= gain
            flag_tot *= flag

        if 'p' in caltype:
            fig = plt.figure(figsize=(ncol / nrow * 10., 8.))
            gs = gridspec.GridSpec(nrow, ncol, left=0.08, right=0.98, wspace=0.3, hspace=0.5)
            for i in range(nant):
                ax = fig.add_subplot(gs[i])
                ax.plot(spws, np.angle(mgain[0, :, i], deg=True), 'o', ms=markersize, color='b')
                ax.plot(spws, np.angle(mgain[1, :, i], deg=True), 'o', ms=markersize, color='r')
                if i / ncol >= nrow - 1:
                    ax.set_xlabel('SPW ID')
                if i % ncol == 0:
                    ax.set_ylabel('Phase (Deg)')
                ax.set_xlim([0, nspw_ms - 1])
                ax.set_ylim([-180, 180])
                ax.set_title('Ant {0}'.format(i))
            plt.suptitle("Phase solutions in gain table {}".format(os.path.basename(slftb)))
            plt.savefig(slftb + '.pha.png')
            plt.close()

        if 'a' in caltype:
            fig = plt.figure(figsize=(ncol / nrow * 10., 8.))
            gs = gridspec.GridSpec(nrow, ncol, left=0.08, right=0.98, wspace=0.3, hspace=0.5)
            for i in range(nant):
                ax = fig.add_subplot(gs[i])
                ax.plot(spws, np.abs(mgain[0, :, i]), 'o', ms=markersize, color='b')
                ax.plot(spws, np.abs(mgain[1, :, i]), 'o', ms=markersize, color='r')
                if i / ncol >= nrow - 1:
                    ax.set_xlabel('SPW ID')
                if i % ncol == 0:
                    ax.set_ylabel('Amplitude')
                ax.set_xlim([0, nspw_ms - 1])
                ax.set_ylim([0.1, 3])
                ax.set_title('Ant {0}'.format(i))
            plt.suptitle("Amplitude solutions in gain table {}".format(os.path.basename(slftb)))
            plt.savefig(slftb + '.amp.png')
            plt.close()

        gains.append(gain)
        mgains.append(mgain)
        flags.append(flag)
        errs.append(err)
        snrs.append(snr)

    if nslftb > 1 and docombine:
        mgain_tot = np.ma.masked_array(gain_tot, mask=flag_tot)
        fig = plt.figure(figsize=(ncol / nrow * 10., 8.))
        gs = gridspec.GridSpec(nrow, ncol, left=0.08, right=0.98, wspace=0.3, hspace=0.5)
        for i in range(nant):
            ax = fig.add_subplot(gs[i])
            ax.plot(spws, np.angle(mgain_tot[0, :, i], deg=True), 'o', ms=markersize, color='b')
            ax.plot(spws, np.angle(mgain_tot[1, :, i], deg=True), 'o', ms=markersize, color='r')
            if i / ncol >= nrow - 1:
                ax.set_xlabel('SPW ID')
            if i % ncol == 0:
                ax.set_ylabel('Phase (Deg)')
            ax.set_xlim([0, nspw_ms - 1])
            ax.set_ylim([-180, 180])
            ax.set_title('Ant {0}'.format(i))
        plt.suptitle("Combined Phases")
        plt.savefig(slftbs[0] + '-' + os.path.basename(slftbs[-1]) + '.pha.comb.png')
        plt.close()

        fig = plt.figure(figsize=(ncol / nrow * 10., 8.))
        gs = gridspec.GridSpec(nrow, ncol, left=0.08, right=0.98, wspace=0.3, hspace=0.5)
        for i in range(nant):
            ax = fig.add_subplot(gs[i])
            ax.plot(spws, np.abs(mgain_tot[0, :, i]), 'o', ms=markersize, color='b')
            ax.plot(spws, np.abs(mgain_tot[1, :, i]), 'o', ms=markersize, color='r')
            if i / ncol >= nrow - 1:
                ax.set_xlabel('SPW ID')
            if i % ncol == 0:
                ax.set_ylabel('Amplitude')
            ax.set_xlim([0, nspw_ms - 1])
            ax.set_ylim([0.1, 3])
            ax.set_title('Ant {0}'.format(i))
        plt.suptitle("Combined Amplitudes")
        plt.savefig(slftbs[0] + '-' + os.path.basename(slftbs[-1]) + '.amp.comb.png')
        plt.close()

    return 1


# ####################### End of function blocks ##############################################
################################################################################################


# ############################### Step 0: Initial Definitions #####################################
# ========== declare input visibility in CASA measurement set format ==========
ms_in = 'IDB20170821201800-202300.4s.corrected.ms'

# define spectral windows used for self-calibration
spws_slf = ['3', '6', '9', '12', '15', '18', '21', '24', '27']
# Uncomment the following to do all of them (spw 0 does not have prior calibration)
# spws_slf = [str(s+1) for s in range(30)]

# ============ Prior definitions for EOVSA data ==============
# define polarization and antennas to self-calibration (no need to change for current EOVSA)
stokes = 'XX'  # polarization to be used for self-calibration
antennas = '0~12'  # antennas numbers used

# =========== TASK HANDLERS =============
dofindflare = 0  # try to find flare time and phase center?

# ============ declaring the working directories (current directory assumed. change as necessary) ============
workdir = os.getcwd() + '/'  # main working directory. Using current directory in this example
slfcaldir = workdir + 'slfcal/'  # place to put all selfcalibration products
imagedir = slfcaldir + 'images_slfcal/'  # place to put all selfcalibration images
maskdir = slfcaldir + 'masks/'  # place to put clean masks
imagedir_slfcaled = slfcaldir + 'images_slfcaled/'  # place to put final self-calibrated images
caltbdir = slfcaldir + 'caltbs/'  # place to put calibration tables
# make these directories if they do not already exist
dirs = [workdir, slfcaldir, imagedir, maskdir, imagedir_slfcaled, caltbdir]
for d in dirs:
    if not os.path.exists(d):
        os.makedirs(d)

# ################################ Step 1: Examine data for flare timing and location ##################
if dofindflare:
    # First, make a cross-power dynamic spectrum by doing median over some appropriate uv range
    d = dspec.Dspec(ms_in, uvrange='0.2~0.8km', usetbtool=True, domedian=True)
    d.plot()
    d.tofits(ms_in + '.dspec.fits')

    # From the previous step, take an appropriate time range for doing self-calibration.
    # Here I select an 8-s piece around the flare peak.
    trange = '2017/08/21/20:21:10~2017/08/21/20:21:18'
    # Do a full-disk imaging to find a new phase center and an appropriate image size.
    qlookplot.qlookplot(ms_in, timerange=trange, spw='3GHz~10GHz', imsize=[512], cell=['5.0arcsec'],
                        niter=100, uvrange='>500lambda', plotaia=False, stokes=stokes)

# Take an appropriate time range for doing self-calibration. Here I select 10-s piece around the flare peak.
trange = '2017/08/21/20:21:10~2017/08/21/20:21:18'
# Using the image it produced, define a new phase center of the image
xycen = [375., 45.]  # in solar coordinates
# Also define the field of view in arcsec
fov = [256., 256.]
# Convert to RA & DEC coordinates (will be used in TCLEAN)
phasecenter, midtime = hf.calc_phasecenter_from_solxy(ms_in, timerange=trange, xycen=xycen,
                                                      usemsphacenter=False)


# ############################ Step 2: Split a short period for self-calibration #################
msinfo = hf.read_msinfo(ms_in, verbose=True)
# output self-calibrated visibility
ms_slfcaled = workdir + os.path.basename(ms_in).replace('corrected', 'slfcaled')

# The following does the split
slfcalms = slfcaldir + 'slfcalms.slfcal'
slfcaledms = slfcaldir + 'slfcalms.slfcaled'
if not os.path.exists(slfcalms):
    print('splitting a small piece for ', trange)
    split(vis=ms_in, outputvis=slfcalms, datacolumn='data', timerange=trange, correlation=stokes, antenna=antennas)


# ################################# Step 3: Main self-calibration loop #############################
# More control parameters for self-calibration
# define ranges of spectral windows you wish to draw masks. Need to change on case-by-case basis
spwran_masks = ['1~5', '6~12', '13~20', '21~30']
half_wid = 2  # how many nearby spws (+- to the spw of choice) to generate model image for the first round of selfcal

# Optional. No need to change in usual cases
maxnround = 10  # maximum rounds of self-calibration allowed
npix = 512
uvrange = '>500lambda'
robust = 1.0
sbeam_1GHz = 80.
refantenna = '0'
sclfactor = 1.  # this applies to VLA images only (with 20 dB attenuators)

# Optional. Consider only this range of spws for imaging/generating models
spw_min = 1  # first spw for consideration
spw_max = 30  # last spw for consideration

os.system('rm -rf ' + imagedir + '*')
os.system('rm -rf ' + caltbdir + '*')
# first step: make a mock caltable for the entire database
print('Processing ' + trange)
msmd.open(slfcalms)
nspw = msmd.nspw()
cfreqsghz = [msmd.meanfreq(s, unit='GHz') for s in range(nspw)]
msmd.done()
spws_ms = [str(s) for s in np.arange(nspw)]

calprefix = caltbdir + 'slf'
imgprefix = imagedir + 'slf'
# starting beam size at 1.0 GHz in arcsec
strtmp = [m.replace(':', '') for m in trange.split('~')]
timestr = 't' + strtmp[0] + '-' + strtmp[1]

# make initial images without self-calibration
n = 0
print('==========Doing initial imaging==========')
clearcal(slfcalms)
prompt3 = input('Wish to draw masks? ')
maskfile = slfcaldir + 'masks.p'
if prompt3.lower() == 'y':
    maskfile = draw_masks(slfcalms, spwrans_mask=spwran_masks, maskdir=maskdir,
                          maskfile=maskfile, restoringbeam=['20arcsec'], timerange=trange,
                          phasecenter=phasecenter, npix=npix, gain=0.05, uvrange=uvrange)
    masks = pickle.load(open(maskfile, 'rb'))
if prompt3.lower() == 'n':
    print('Chose not to draw new masks. Looking for existing ones.')
    if os.path.exists(slfcaldir + 'masks.p'):
        masks = pickle.load(open(slfcaldir + 'masks.p', 'rb'))
    if not os.path.exists(slfcaldir + 'masks.p'):
        print('masks do not exist. Use default mask')
        masks = []

figname, fitsfiles = make_images(slfcalms, msinfo=msinfo, spws=spws_slf, spwrans_mask=spwran_masks,
                                 usemask=True, maskfile=maskfile, imgprefix=imgprefix, round=n,
                                 trange=trange, xycen=xycen, fov=fov,
                                 uvrange=uvrange, antennas='', npix=npix, cell=['1arcsec'],
                                 niter=500, gain=0.05, robust=1.0, sbeam_1GHz=sbeam_1GHz, nrow=3,
                                 usemsphacenter=False, observatory='EOVSA',sclfactor=sclfactor)
print('Resulting image is {}'.format(figname))
slftbs = []
caltypes = []
niters = []

# Question on whether phase or amplitude selfcalibration
prompt2 = input('Phase only (p), amplitude only (a), or amplitude & phase (ap)? ')
while prompt2.lower() != 'p' and prompt2.lower() != 'a' and prompt2.lower() != 'ap':
    print('Your input is {}. Your input has to be "p", "a", or "ap"'.format(prompt2))
    prompt2 = input('Phase only (p), amplitude only (a), or amplitude & phase (ap)? ')
if prompt2.lower() == 'p':
    caltype = 'p'
    print('Doing phase-only calibration')
if prompt2.lower() == 'a':
    caltype = 'a'
    print('Doing amplitude-only calibration')
if prompt2.lower() == 'ap':
    caltype = 'ap'
    print('Doing amplitude & phase calibration')
caltypes.append(caltype)
# Question on how many iterations
while True:
    try:
        niter = int(input('How many iterations (e.g., 100)? '))
    except ValueError:
        print('Your input has to be an integer number. Try again.')
    else:
        break
print('Using {0:d} iterations for this self-calibration round'.format(niter))
niters.append(niter)

while n < maxnround:
    # Name of the self-calibration gain table, round is 1 based
    slfcal_tb_g = calprefix + '.G' + str(n + 1)

    solv_success = False
    for s, sp in enumerate(spws_slf):
        print('Self-calibrating spw: ' + sp)
        cfreqghz = cfreqsghz[int(sp)]
        # setting restoring beam size (not very useful for selfcal anyway, but just to see the results)
        bm = max(sbeam_1GHz / cfreqghz, 4.)
        print('Beam size {0:.1f}" at ref freq {1:.2f} GHz'.format(bm, cfreqghz))
        slfcal_img = imgprefix + '.spw' + sp.zfill(2) + '.slfcal' + str(n) + '.tmp'
        # only the first and second round uses nearby spws for getting initial model
        if n < 1:
            if not spw_min:
                spw_min = int(spws_ms[0])
            if not spw_max:
                spw_min = int(spws_ms[-1])
            spbg = int(max(int(sp) - half_wid, spw_min))
            sped = int(min(int(sp) + half_wid, spw_max))
            spwran = str(spbg) + '~' + str(sped)
            print('using spw {0:s} as model'.format(spwran))
        else:
            spwran = sp
            print('using spw {0:s} as model'.format(spwran))
        foundmask = False
        if len(masks) > 0:
            for m, spwran_mask in enumerate(spwran_masks):
                sbg, sed = spwran_mask.split('~')
                if (int(sp) >= int(sbg)) and (int(sp) <= int(sed)):
                    mask = masks[m]
                    print('using mask {0:s}'.format(os.path.basename(mask)))
                    foundmask = True
            if not foundmask:
                print('mask not defined for {}. Do use any masks.'.format(sp))
                mask = ''
        else:
            mask = ''
        try:
            tclean(vis=slfcalms, antenna=antennas, imagename=slfcal_img, uvrange=uvrange, spw=spwran,
                   specmode='mfs', timerange=trange, imsize=[npix], cell=['1arcsec'], niter=niter,
                   gain=0.05, stokes=stokes, weighting='briggs', robust=robust, phasecenter=phasecenter, mask=mask,
                   restoringbeam=[str(bm) + 'arcsec'], pbcor=False, interactive=False, savemodel='modelcolumn')

            clnjunks = ['.mask', '.flux', '.model', '.psf', '.residual', '.image', '.pb', '.image.pbcor', '.sumwt']
            for clnjunk in clnjunks:
                if os.path.exists(slfcal_img + clnjunk):
                    os.system('rm -rf ' + slfcal_img + clnjunk)
        except:
            print('error in cleaning spw: ' + sp)
            print('using nearby spws for initial model')
            sp_e = min(int(sp) + 2, int(spws_ms[-1]))
            sp_i = max(int(sp) - 2, int(spws_ms[0]))
            sp_ = str(sp_i) + '~' + str(sp_e)
            try:
                tclean(vis=slfcalms, antenna=antennas, imagename=slfcal_img, uvrange=uvrange, spw=sp_,
                       specmode='mfs', timerange=trange, imsize=[npix], cell=['1arcsec'], niter=niter,
                       gain=0.05, stokes=stokes, weighting='briggs', robust=robust, phasecenter=phasecenter,
                       mask=mask, restoringbeam=[str(bm) + 'arcsec'],
                       pbcor=False, interactive=False, savemodel='modelcolumn')
                print('using spw {0:s} as model'.format(sp_))
            except:
                print('still not successful. abort...')
                solv_success = False
                continue

        # Check what the calibration type is. If it involves amplitude, need to use the same uvrange at used in
        # generating the model image.
        if caltype == 'p':
            uvrange_for_gaincal = ''
        else:
            uvrange_for_gaincal = uvrange

        if s == 0 or not solv_success:
            append = False
        else:
            append = True
        try:
            # Tested (in CASA 6.5) that flags in gaintables being applied on the fly do not affect the solution
            # (equivalent to applymode='calonly')
            gaincal(vis=slfcalms, refant=refantenna, antenna=antennas, caltable=slfcal_tb_g, spw=sp,
                    uvrange=uvrange_for_gaincal, gaintable=slftbs, selectdata=True, timerange=trange,
                    solint='inf', gaintype='G', calmode=caltype, combine='', minblperant=3, minsnr=2,
                    append=append)
        except:
            print('No solution found in spw: ' + sp)
            solv_success = False
            continue

        if os.path.exists(slfcal_tb_g):
            solv_success = True
        else:
            solv_success = False

    if os.path.exists(slfcal_tb_g):
        slftbs.append(slfcal_tb_g)
        clearcal(slfcalms)
        delmod(slfcalms)
        applycal(vis=slfcalms, gaintable=slftbs, spw=','.join(spws_slf), selectdata=True,
                 antenna=antennas, interp='nearest', flagbackup=False, applymode='calonly', calwt=False)
        res = plt_slftable(slfcalms, slfcal_tb_g, caltype, docombine=False)

    if solv_success:
        # make images after this self-calibration round
        figname, fitsfiles = make_images(slfcalms, msinfo=msinfo, spws=spws_slf, spwrans_mask=spwran_masks,
                                         usemask=False, imgprefix=imgprefix, round=n + 1,
                                         trange=trange, xycen=xycen, fov=fov, uvrange=uvrange, antennas='',
                                         npix=npix,
                                         cell=['1arcsec'], niter=500, gain=0.1, robust=1.0, sbeam_1GHz=sbeam_1GHz,
                                         nrow=3, usemsphacenter=False, observatory='EOVSA', sclfactor=sclfactor)
        print('Resulting image is {}'.format(figname))
        print('Check the resulting image {0} and calibration table plot in {1}.'.format(figname,
                                                                                        os.path.dirname(
                                                                                            slfcal_tb_g)))

    prompt = input('Continue selfcal? (y or n) ')
    while prompt.lower() != 'y' and prompt.lower() != 'n':
        print('Your input is {}. Your input has to be "y" or "n"'.format(prompt))
        prompt = input('Continue selfcal? (y or n) ')
    if prompt.lower() == 'n':
        if os.path.exists(slfcaledms):
            os.system('rm -rf ' + slfcaledms)
        split(slfcalms, slfcaledms, datacolumn='corrected')
        print('Final self-calibrated ms is {0:s}'.format(slfcaledms))
        success = plt_slftable(slfcalms, slftbs, caltypes, docombine=True)
        outfits = '{0:s}/slfcaled_image_comb.fits'.format(imagedir_slfcaled)
        print('Final self-calibrated image is {}'.format(outfits))
        if len(fitsfiles) > 1:
            fitsutils.fits_wrap_spwX(fitsfiles, outfits)
        break
    if prompt.lower() == 'y':
        n = n + 1
        if n < maxnround:
            print('Continue to the No. {0:d} round'.format(n + 1))
            # Question on whether phase or amplitude selfcalibration
            prompt2 = input('Phase only (p), amplitude only (a), or amplitude & phase (ap)? ')
            while prompt2.lower() != 'p' and prompt2.lower() != 'a' and prompt2.lower() != 'ap':
                print('Your input is {}. Your input has to be "p", "a", or "ap"'.format(prompt2))
                prompt2 = input('Phase only (p), amplitude only (a), or amplitude & phase (ap)? ')
            if prompt2.lower() == 'p':
                caltype = 'p'
                print('Doing phase-only calibration')
            if prompt2.lower() == 'a':
                caltype = 'a'
                print('Doing amplitude-only calibration')
            if prompt2.lower() == 'ap':
                caltype = 'ap'
                print('Doing amplitude & phase calibration')
            caltypes.append(caltype)
            # Question on making new masks
            prompt3 = input('Wish to update masks? (y or n) ')
            while prompt3.lower() != 'y' and prompt3.lower() != 'n':
                print('Your input is {}. Your input has to be "y" or "n"'.format(prompt3))
                prompt3 = input('Wish to update masks? (y or n) ')
            if prompt3.lower() == 'y':
                maskfile = draw_masks(slfcalms, spwrans_mask=spwran_masks, maskdir=maskdir,
                                      maskfile=maskfile, restoringbeam=['20arcsec'], timerange=trange,
                                      phasecenter=phasecenter, npix=npix, gain=0.05, uvrange=uvrange)
                masks = pickle.load(open(maskfile, 'rb'))
            else:
                print('Chose not to draw new masks. Looking for existing ones.')
                if os.path.exists(slfcaldir + 'masks.p'):
                    masks = pickle.load(open(slfcaldir + 'masks.p', 'rb'))
                if not os.path.exists(slfcaldir + 'masks.p'):
                    print('masks do not exist. Use default mask')
                    masks = []
            # Question on how many iterations
            while True:
                try:
                    niter = int(input('How many iterations (e.g., 100)? '))
                except ValueError:
                    print('Your input has to be an integer number. Try again.')
                else:
                    break
            print('Using {0:d} iterations for this self-calibration round'.format(niter))
            niters.append(niter)
        else:
            print('Hitting the max rounds of self-calibration of {0:d}'.format(n))
            print('Splitting final calibrated ms is {0:s}'.format(slfcaledms))
            if os.path.exists(slfcaledms):
                os.system('rm -rf ' + slfcaledms)
            split(slfcalms, slfcaledms, datacolumn='corrected')
            try:
                success = plt_slftable(slfcalms, slftbs, caltypes, docombine=True)
            except:
                print('somehow combining the gain tables from all rounds failed. Proceed without doing the plot.')
            outfits = '{0:s}/slfcaled_image_comb.fits'.format(imagedir_slfcaled)
            print('Final self-calibrated image is {}'.format(outfits))
            if len(fitsfiles) > 1:
                fitsutils.fits_wrap_spwX(fitsfiles, outfits)

# ####################### Step 4: Apply self-calibration tables to the original visibility dataset ####################
prompt4 = input('Do you want to apply the self-calibration tables to the original ms? (y or n) ')
while prompt4.lower() != 'y' and prompt4.lower() != 'n':
    print('Your input is {}. Your input has to be "y" or "n"'.format(prompt4))
    prompt4 = input('Do you want to apply the self-calibration tables to the original ms? (y or n) ')
if prompt4.lower() == 'y':
    clearcal(ms_in)
    applycal(vis=ms_in, gaintable=slftbs, spw=','.join(spws_slf), selectdata=True,
             antenna=antennas, interp='nearest', flagbackup=False, applymode='calonly', calwt=False)
    if os.path.exists(ms_slfcaled):
        print('Found an existing self-calibrated dataset.')
        prompt5 = input('Wish to overwrite this dataset? (y or n) ')
        while prompt5.lower() != 'y' and prompt5.lower() != 'n':
            print('Your input is {}. Your input has to be "y" or "n"'.format(prompt5))
            prompt5 = input('Wish to overwrite this dataset? (y or n) ')
        if prompt5.lower() == 'y':
            os.system('rm -rf ' + ms_slfcaled)
            ms_out = ms_slfcaled
            print('Splitting self-calibrated dataset as {}'.format(ms_out))
        else:
            ms_out = ms_slfcaled + '.copy'
    else:
        ms_out = ms_slfcaled
    print('Splitting self-calibrated dataset as {}'.format(ms_out))
    split(ms_in, ms_out, spw=','.join(spws_slf), correlation=stokes, antenna=antennas, datacolumn='corrected')
else:
    print('Chose not to apply self-calibration tables. You can apply them later using applycal().')
    print('Self-calibration tables are:', slftbs)





##################### generated by xml-casa (v2) from pmaxfit.xml ###################
##################### b3be0c8ff5a66315d1c40770a10aa871 ##############################
from __future__ import absolute_import
import numpy
from casatools.typecheck import CasaValidator as _val_ctor
_pc = _val_ctor( )
from casatools.coercetype import coerce as _coerce
from .private.task_pmaxfit import pmaxfit as _pmaxfit_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log

class _pmaxfit:
    """
    pmaxfit ---- Find maximum and do parabolic fit in the sky

    
    PARAMETER SUMMARY
    imagename        Name of the input image
    box              Rectangular region(s) to select in direction plane. See "help par.box"
    for details. Default is to use the entire direction plane.
    eg "100, 120, 200, 220, 300, 300, 400, 400" to use two boxes.
    
    OVERVIEW
    This application finds the pixel with the maximum value in the region, and then uses function
    findsources to generate a Componentlist with one component.
    
    The method returns a dictionary with fours keys, 'succeeded', 'timestamps', 'imagenames'
    and 'outputs'. The value of 'outputs' is a dictionary representing
    a component list reflecting the fit results over multiple channels.
    Both the 'outputs' dictionaries can be read into a component list tool (default tool is named cl)
    using the fromrecord() method for easier inspection using tool methods, eg
    
    FITTING OVER MULTIPLE CHANNELS
    
    For fitting over multiple channels, the result of the previous successful fit is used as
    the estimate for the next channel. The number of gaussians fit cannot be varied on a channel
    by channel basis. Thus the variation of source structure should be reasonably smooth in
    frequency to produce reliable fit results.
    

    --------- parameter descriptions ---------------------------------------------

    imagefiles A list of the input images
    ncpu       Number of cpu cores to use
    box        Rectangular region(s) to select in direction plane. See "help par.box" for details. Default is to use the entire direction plane.
    width      Half-width of fit grid
    [1;42mRETURNS[1;m       void

    --------- examples -----------------------------------------------------------

    
    
    
    EXAMPLE:
    
    Here is how one might fit two gaussians to multiple channels of a cube using the fit
    from the previous channel as the initial estimate for the next. It also illustrates
    how one can specify a region in the associated continuum image as the region to use
    as the fit for the channel.
    
    
    begin{verbatim}
    default pmaxfit
    imagename = "co_cube.im"
    # specify region using region from continuum
    box = "100,120,200,200"
    pmaxfit()
    


    """

    _info_group_ = """analysis"""
    _info_desc_ = """Find maximum and do parabolic fit in the sky"""

    def __call__( self, imagefiles=[  ], ncpu=int(8), box='', width=int(5) ):
        schema = {'imagefiles': {'type': 'cReqPathVec', 'coerce': [_coerce.to_list,_coerce.expand_pathvec]}, 'ncpu': {'type': 'cInt'}, 'box': {'type': 'cStr', 'coerce': _coerce.to_str}, 'width': {'type': 'cInt'}}
        doc = {'imagefiles': imagefiles, 'ncpu': ncpu, 'box': box, 'width': width}
        assert _pc.validate(doc,schema), str(_pc.errors)
        _logging_state_ = _start_log( 'pmaxfit', [ 'imagefiles=' + repr(_pc.document['imagefiles']), 'ncpu=' + repr(_pc.document['ncpu']), 'box=' + repr(_pc.document['box']), 'width=' + repr(_pc.document['width']) ] )
        return _end_log( _logging_state_, 'pmaxfit', _pmaxfit_t( _pc.document['imagefiles'], _pc.document['ncpu'], _pc.document['box'], _pc.document['width'] ) )

pmaxfit = _pmaxfit( )




import os
import h5py
import numpy as np
import datetime
from astropy.time import Time
import itertools
from astropy import units as u
import pandas as pd
from astropy.coordinates import SkyCoord, EarthLocation, get_body, AltAz

def rebin1d(arr, new_len):
    shape = (new_len, len(arr) // new_len)
    return arr.reshape(shape).mean(1)


def rebin2d(arr, new_shape):
    shape = (new_shape[0], arr.shape[0] // new_shape[0],
             new_shape[1], arr.shape[1] // new_shape[1])
    return arr.reshape(shape).mean(-1).mean(1)


def timestamp_to_mjd(times):
    # This is from Ivey Davis's BeamTools.py
    t_flat = np.array(list(itertools.chain(*times)))
    ts_inds = np.linspace(0, len(times)-1, len(times), dtype  = int)*2
    other_inds = ts_inds + 1
    ts = Time(t_flat[ts_inds],format = 'unix') + t_flat[other_inds]* u.s
    ts = ts.mjd
    return ts


def read_data(filename, stokes='I', timerange=[], freqrange=[], timebin=1, freqbin=1, verbose=True, 
            flux_factor_file=None, bkg_file=None,  do_pb_correction=False, 
            flux_factor_calfac_x = None, flux_factor_calfac_y = None, bkg_flux_arr = None):
    '''
    :param filename: name of the OVRO-LWA hdf5 beamforming file; 
              This can be a string (single file) or a list of strings (multiple files)
    :param stokes: currently supporting 'XX', 'YY', 'I', 'Q', 'U', 'V', 'IV' 
    :param timerange: list of [start_time, end_time], start_time and end_time should be recognized by astropy.time.Time
            e.g., ['2023-09-22T18:00', '2023-09-22T18:10']
    :param freqrange: list of [start_frequency, end_frequency] in MHz. Example: [23, 82]
    :param timebin: number to bin in time
    :param freqbin: number to bin in frequency
    :param verbose: if True, print extra information
    :param flux_factor_file: Path to the csv file that contains the flux correction factors
    :param bkg_file: Path to the csv file that contains the raw background (off-Sun scan) measurements (before scaling); 
            currently it only contains Stokes I.
    :param do_pb_correction: if True, apply primary beam correction. Currently only use the analytical form in Stokes I only.
    :param flux_factor_calfac_x: user input correction factor for the X polarization
    :param flux_factor_calfac_y: user input correction factor for the Y polarization
    :param bkg_flux_arr: user input background flux in Jy
    '''
    # Check the input filename
    if type(filename) == str:
        filename = [filename]

    obs = EarthLocation.of_site('ovro')
    filelist = []
    for ll in filename:
        if not os.path.exists(ll):
            print('{} does not exist. Skip this one.'.format(ll))
        else:
            filelist.append(ll)

    if not filelist:
        print('I cannot find any file. Abort.')
        return False
    else:
        filelist.sort()
    
    n0 = 0 # this is the index for the first file being read successfully
    firstset_read = False 
    for n, file in enumerate(filelist):
        if verbose:
            print('Processing {0:d} of {1:d} files'.format(n+1, len(filelist)))
        try:
            data = h5py.File(file, 'r',swmr=True)
        except:
            print('Cannot read {0:s}. Skip this file.'.format(file))
            continue
        freqs = data['Observation1']['Tuning1']['freq'][:]
        ts = data['Observation1']['time'][:]
        # The following line works the same way as timestamp_to_mjd(), but a bit too slow
        # times_mjd = np.array([(Time(t[0], format='unix') + TimeDelta(t[1], format='sec')).mjd for t in ts])
        times_mjd = timestamp_to_mjd(ts)
        idx0, = np.where(times_mjd > 50000.) # filter out those prior to 1995 (obviously wrong for OVRO-LWA)

        # read the flux factors file if provided
        if not (flux_factor_file is None): 
            try:
                out = pd.read_csv(flux_factor_file)
                calfac_x = np.array(out['calfac_x'])
                calfac_y = np.array(out['calfac_y'])
            except:
                print('Failed in reading the flux factor csv file. Setting correction factors to unity.')
        else:
            print('Flux factor csv file does not exist. Setting correction factors to unity.')
            calfac_x = np.ones_like(freqs)
            calfac_y = np.ones_like(freqs)

        if not (flux_factor_calfac_x is None) and not (flux_factor_calfac_y is None):
            # user input correction factor
            calfac_x = calfac_x*flux_factor_calfac_x
            calfac_y = calfac_y*flux_factor_calfac_y
        
        if not (bkg_flux_arr is None):
            # add the user input background flux
            bkg_flux += bkg_flux_arr

        # read background flux file if provided
        if not (bkg_file is None): 
            try:
                out = pd.read_csv(bkg_file)
                bkg_flux = out['bkg_flux']
                print('Using the provided raw aackground flux csv file.')
            except:
                print('Failed in reading the background flux csv file. Setting background flux to zero.')
        else:
            print('No background csv file provided. Setting background flux to zero.')
            bkg_flux = np.zeros_like(freqs)

        if verbose:
            print('Data time range is from {0:s} to {1:s}'.format(Time(times_mjd[idx0][0], format='mjd').isot, 
                Time(times_mjd[idx0][-1], format='mjd').isot))
            print('Data has {0:d} time stamps and {1:d} frequency channels'.format(len(times_mjd[idx0]), len(freqs)))

        # Select time range
        if len(timerange) > 0:
            try:
                timerange_obj = Time(timerange)
                # Take the larger value of the supplied start time and the first time stamp of the data
                t0 = max(timerange_obj[0].mjd, min(times_mjd[idx0]))
                # Take the smaller value of the supplied end time and the last time stamp of the data
                t1 = min(timerange_obj[1].mjd, max(times_mjd[idx0]))
                ti0 = np.argmin(np.abs(times_mjd - t0))
                ti1 = np.argmin(np.abs(times_mjd - t1)) 
                if ti1 - ti0 < timebin:
                    print('Selected number of time samples {0:d} is less than the timebin {1:d}. Skip this file.'.format(ti1-ti0, timebin))
                    if not firstset_read:
                        n0 += 1
                    continue
                if verbose:
                    print('Selected time range is from {0:s} to {1:s}'.format(Time(times_mjd[ti0], format='mjd').isot, 
                                                                              Time(times_mjd[ti1], format='mjd').isot))
            except:
                print('timerange not parsed correctly. Use the full range in the data.')
                ti0 = 0
                ti1 = len(times_mjd) 
        else:
            ti0=0
            ti1=len(times_mjd)

        firstset_read = True
        times_mjd = times_mjd[ti0:ti1] 

        # Select frequency range
        if len(freqrange) > 0:
            if type(freqrange) == list:
                try:
                    f0 = freqrange[0]
                    f1 = freqrange[1]
                    if f0 > 100. or f1 > 100.:
                        # I am assuming input frequency range is in Hz
                        print('Input frequency range is greater than 100. Assuming unit in Hz.')
                    else:
                        # I am assuming input frequency range is in MHz
                        print('Input frequency range is less than 100. Assuming unit in MHz.')
                        f0 *= 1e6
                        f1 *= 1e6
                    fi0 = np.argmin(np.abs(freqs - f0))
                    fi1 = np.argmin(np.abs(freqs - f1)) 
                except:
                    print('freqrange not parsed correctly. Use the full range.')
                    fi0 = 0
                    fi1 = len(freqs) 
        else:
            fi0=0
            fi1=len(freqs)

        freqs = freqs[fi0:fi1] 

        # select stokes
        stokes_valid = ['XX', 'YY', 'I', 'Q', 'U', 'V', 'IV']
        if verbose:
            print('Reading dynamic spectrum for stokes {0:s}'.format(stokes))

        if stokes not in stokes_valid:
            raise Exception("Provided Stokes {0:s} is not in 'XX, YY, RR, LL, I, Q, U, V'".format(stokes))
        if stokes.upper() == 'XX':
            spec = data['Observation1']['Tuning1']['XX'][ti0:ti1, fi0:fi1] / calfac_x[None, fi0:fi1]
            stokes_out = ['XX']
        if stokes.upper() == 'YY':
            spec = data['Observation1']['Tuning1']['YY'][ti0:ti1, fi0:fi1] / calfac_y[None, fi0:fi1]
            stokes_out = ['YY']
        if stokes.upper() == 'I' or stokes.upper() == 'IV':
            spec_I = (data['Observation1']['Tuning1']['XX'][ti0:ti1, fi0:fi1] / calfac_x[None, fi0:fi1] + \
                   data['Observation1']['Tuning1']['YY'][ti0:ti1, fi0:fi1] / calfac_y[None, fi0:fi1]) / 2. - \
                   bkg_flux[None, fi0:fi1] / ((calfac_x[None, fi0:fi1] + calfac_y[None, fi0:fi1]) / 2.) 
            if verbose:
                print('Median of the subtracted background flux (Jy)', np.median(bkg_flux[None, fi0:fi1] / ((calfac_x[None, fi0:fi1] + calfac_y[None, fi0:fi1]) / 2.)))
                print('RMS of the subtracted background flux (Jy)', np.std(bkg_flux[None, fi0:fi1] / ((calfac_x[None, fi0:fi1] + calfac_y[None, fi0:fi1]) / 2.)))
            if do_pb_correction:
                pbfacs = np.ones_like(times_mjd)
                t0 = times_mjd[0]
                t1 = times_mjd[-1]
                if t1-t0 > 5./1440.:
                    nstep = int((t1-t0)/(5./1440.))
                    ts_ref = np.linspace(t0, t1, nstep)
                    pbfacs_ref = np.ones_like(ts_ref) 
                    print('Duration of the file is {0:.1f} hours, interpolating into {1:d} steps.'.format((t1-t0)*24., nstep))
                    for i, t_ref in enumerate(ts_ref):
                        sun_loc = get_body('sun', Time(t_ref, format='mjd'), location=obs)
                        alt = sun_loc.transform_to(AltAz(obstime=Time(t_ref, format='mjd'), location=obs)).alt.radian
                        if np.degrees(alt) > 5.:
                            pbfacs_ref[i]=np.sin(alt)**1.6
                        else:
                            print('Warning! Calculated solar altitude is lower than 5 degrees. Something is wrong with the data (non-solar)?') 
                            pbfacs_ref[i]=np.sin(np.radians(5.))**1.6

                    pbfacs = np.interp(times_mjd, ts_ref, pbfacs_ref)
                else:
                    print('Duration of the file is {0:.1f} minutes, no interpolation will be done.'.format((t1-t0)*24.*60.))
                    t_ref = (t0+t1)/2.
                    sun_loc = get_body('sun', Time(t_ref, format='mjd'), location=obs)
                    alt = sun_loc.transform_to(AltAz(obstime=Time(t_ref, format='mjd'), location=obs)).alt.radian
                    if np.degrees(alt) > 5.:
                        pbfacs *= np.sin(alt)**1.6
                    else:
                        print('Warning! Calculated solar altitude is lower than 5 degrees. Something is wrong with the data (non-solar)?') 
                        pbfacs *= np.sin(np.radians(5.))**1.6
                spec_I /= pbfacs[:, None]

            if stokes.upper() == 'IV':
                spec_V = data['Observation1']['Tuning1']['XY_imag'][ti0:ti1, fi0:fi1] / ((calfac_x[None, fi0:fi1] + calfac_y[None, fi0:fi1]) / 2.)
                spec = np.stack((spec_I, spec_V), axis=2)
                stokes_out = ['I', 'V']
            else:
                spec = spec_I
                stokes_out = ['I'] 
        if stokes == 'V':
            spec = data['Observation1']['Tuning1']['XY_imag'][ti0:ti1, fi0:fi1] / ((calfac_x[None, fi0:fi1] + calfac_y[None, fi0:fi1]) / 2.)
            stokes_out = ['V'] 
        if stokes == 'Q':
            spec = (data['Observation1']['Tuning1']['XX'][ti0:ti1, fi0:fi1] / calfac_x[None, fi0:fi1] - \
                    data['Observation1']['Tuning1']['YY'][ti0:ti1, fi0:fi1] / calfac_y[None, fi0:fi1]) / 2. 
            stokes_out = ['Q'] 
        if stokes == 'U':
            spec = data['Observation1']['Tuning1']['XY_real'][ti0:ti1, fi0:fi1] / ((calfac_x[None, fi0:fi1] + calfac_y[None, fi0:fi1]) / 2.)
            stokes_out = ['U'] 

        idx, = np.where(times_mjd > 50000.) # filter out those prior to 1995 (obviously wrong for OVRO-LWA)
        times_mjd = times_mjd[idx]
        spec = spec[idx]

        if spec.ndim == 2:
            # TODO: for now I have just ignored the rest of the data that falls outside of the whole factor of timebin * nt_new or freqbin * nf_new 
            nt, nf = spec.shape
            npol = 1
            nt_new, nf_new = (nt // timebin, nf // freqbin)
            spec_new = rebin2d(spec[:nt_new*timebin, :nf_new*freqbin], (nt_new, nf_new))
        if spec.ndim == 3:
            nt, nf, npol = spec.shape
            spec_new_ = [] 
            for i in range(npol):
                nt_new, nf_new = (nt // timebin, nf // freqbin)
                spec_ = spec[:, :, i]
                spec_new_.append(rebin2d(spec_[:nt_new*timebin, :nf_new*freqbin], (nt_new, nf_new)))
            spec_new = np.stack(spec_new_, axis=2)
            
        spec_new = np.transpose(spec_new).reshape((npol, 1, nf_new, nt_new)) / 1e4
        times_mjd_new = rebin1d(times_mjd[:nt_new * timebin], nt_new)
        freqs_new = rebin1d(freqs[:nf_new * freqbin], nf_new)
        if firstset_read:
            if n == n0:
                nfreq0 = len(freqs_new)
                freqs_out = freqs_new
                spec_out = spec_new
                times_mjd_out = times_mjd_new
            elif n > n0:
                if len(freqs_new) != nfreq0:
                    print('Something is wrong in concatenating {}'.format(file)) 
                    print('Dimension of the output frequency {0:d} does not match that of the first file {1:d)'.format(len(freqs_new), nfreq0)) 
                    continue
                else:
                    spec_out = np.concatenate((spec_out, spec_new), axis=3)
                    times_mjd_out = np.concatenate((times_mjd_out, times_mjd_new))
        else:
            continue

    if firstset_read:
        if verbose:
            print('Output time range is from {0:s} to {1:s}'.format(Time(times_mjd_out[0], format='mjd').isot, Time(times_mjd_out[-1], format='mjd').isot))
            print('Output data has {0:d} time stamps and {1:d} frequency channels'.format(len(times_mjd_out), len(freqs_out)))
        return spec_out, times_mjd_out, freqs_out, stokes_out, calfac_x, calfac_y, bkg_flux
    else:
        return False
